from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Final, Literal

_ALIAS_DECLARATION_PATTERN: Final[re.Pattern[str]] = re.compile(
    r"\bexport\s+type\s+(?P<name>[A-Za-z_][A-Za-z0-9_]*)\s*=",
    re.MULTILINE,
)
_IDENTIFIER_PATTERN: Final[re.Pattern[str]] = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")
_STRING_QUOTES: Final[frozenset[str]] = frozenset({'"', "'", "`"})
_MIN_STRING_LITERAL_LENGTH: Final[int] = 2


@dataclass(frozen=True, slots=True)
class TsAlias:
    name: str
    kind: Literal["object", "string_union", "identifier_union"]
    properties: dict[str, TsProperty] | None = None
    string_literals: tuple[str, ...] | None = None
    union_members: tuple[str, ...] | None = None


@dataclass(frozen=True, slots=True)
class TsProperty:
    name: str
    optional: bool
    type_expr: str
    inline_object: TsAlias | None = None


@dataclass(slots=True)
class _CommentState:
    in_string: str | None = None
    escaped: bool = False

    def consume_string_char(self, char: str) -> None:
        if self.escaped:
            self.escaped = False
            return

        if char == "\\":
            self.escaped = True
            return

        if char == self.in_string:
            self.in_string = None


@dataclass(slots=True)
class _DepthState:
    brace_depth: int = 0
    bracket_depth: int = 0
    paren_depth: int = 0
    in_string: str | None = None
    escaped: bool = False

    def consume_char(self, char: str) -> None:
        if self.in_string is not None:
            self._consume_string_char(char)
            return

        if char in _STRING_QUOTES:
            self.in_string = char
            return

        self.brace_depth += _depth_delta(char, opening="{", closing="}")
        self.bracket_depth += _depth_delta(char, opening="[", closing="]")
        self.paren_depth += _depth_delta(char, opening="(", closing=")")

    def _consume_string_char(self, char: str) -> None:
        if self.escaped:
            self.escaped = False
            return

        if char == "\\":
            self.escaped = True
            return

        if char == self.in_string:
            self.in_string = None


def parse_exported_type_aliases(ts_source: str) -> dict[str, TsAlias]:
    stripped_source = _strip_comments(ts_source)
    aliases: dict[str, TsAlias] = {}

    for match in _ALIAS_DECLARATION_PATTERN.finditer(stripped_source):
        alias_name = match.group("name")
        alias_rhs, _ = _capture_rhs_until_semicolon(stripped_source, start=match.end())
        aliases[alias_name] = _parse_alias(alias_name, alias_rhs.strip())

    return aliases


def _strip_comments(source: str) -> str:
    result: list[str] = []
    index = 0
    state = _CommentState()

    while index < len(source):
        if state.in_string is not None:
            char = source[index]
            result.append(char)
            state.consume_string_char(char)
            index += 1
            continue

        if _starts_block_comment(source, index):
            index = _skip_block_comment(source, index, result=result)
            continue

        if _starts_line_comment(source, index):
            index = _skip_line_comment(source, index)
            continue

        char = source[index]
        if char in _STRING_QUOTES:
            state.in_string = char
        result.append(char)
        index += 1

    return "".join(result)


def _capture_rhs_until_semicolon(source: str, *, start: int) -> tuple[str, int]:
    state = _DepthState()
    for index in range(start, len(source)):
        char = source[index]
        state.consume_char(char)
        if char == ";" and _is_top_level(state):
            return source[start:index], index + 1

    msg = f"Could not find terminating semicolon for alias at index {start}"
    raise ValueError(msg)


def _parse_alias(alias_name: str, rhs: str) -> TsAlias:
    if rhs.startswith("{"):
        return TsAlias(
            name=alias_name,
            kind="object",
            properties=_parse_object_properties(alias_name, rhs),
        )

    union_tokens = [
        token.strip() for token in _split_top_level(rhs, delimiter="|") if token.strip()
    ]
    if not union_tokens:
        msg = f"Alias {alias_name} has an empty right-hand side"
        raise ValueError(msg)

    string_literals: list[str] = []
    for token in union_tokens:
        parsed = _parse_string_literal(token)
        if parsed is None:
            break
        string_literals.append(parsed)
    else:
        return TsAlias(
            name=alias_name,
            kind="string_union",
            string_literals=tuple(string_literals),
        )

    for token in union_tokens:
        if not _IDENTIFIER_PATTERN.fullmatch(token):
            msg = f"Alias {alias_name} includes unsupported union member: {token!r}"
            raise ValueError(msg)

    return TsAlias(
        name=alias_name,
        kind="identifier_union",
        union_members=tuple(union_tokens),
    )


def _parse_object_properties(alias_name: str, object_expr: str) -> dict[str, TsProperty]:
    object_body = _extract_object_body(alias_name, object_expr)
    raw_properties = _split_top_level(object_body, delimiter=";")

    properties: dict[str, TsProperty] = {}
    for raw_property in raw_properties:
        stripped_property = raw_property.strip()
        if not stripped_property:
            continue

        property_name, optional, type_expr = _parse_property(alias_name, stripped_property)
        inline_object: TsAlias | None = None
        if type_expr.lstrip().startswith("{"):
            inline_object = TsAlias(
                name=f"{alias_name}.{property_name}",
                kind="object",
                properties=_parse_object_properties(f"{alias_name}.{property_name}", type_expr),
            )

        properties[property_name] = TsProperty(
            name=property_name,
            optional=optional,
            type_expr=type_expr,
            inline_object=inline_object,
        )

    return properties


def _extract_object_body(alias_name: str, object_expr: str) -> str:
    stripped_expr = object_expr.strip()
    if not stripped_expr.startswith("{"):
        msg = f"Alias {alias_name} is expected to start with '{{'"
        raise ValueError(msg)

    state = _DepthState()
    closing_index: int | None = None

    for index, char in enumerate(stripped_expr):
        previous_brace_depth = state.brace_depth
        state.consume_char(char)
        if (
            char == "}"
            and previous_brace_depth == 1
            and state.brace_depth == 0
            and state.in_string is None
        ):
            closing_index = index
            break

    if closing_index is None:
        msg = f"Alias {alias_name} object has unbalanced braces"
        raise ValueError(msg)

    trailing_text = stripped_expr[closing_index + 1 :].strip()
    if trailing_text:
        msg = f"Alias {alias_name} object has unsupported trailing expression: {trailing_text!r}"
        raise ValueError(msg)

    return stripped_expr[1:closing_index]


def _parse_property(alias_name: str, raw_property: str) -> tuple[str, bool, str]:
    separator_index = _find_top_level(raw_property, target=":")
    if separator_index == -1:
        msg = f"Alias {alias_name} contains malformed property: {raw_property!r}"
        raise ValueError(msg)

    raw_name = raw_property[:separator_index].strip()
    raw_type_expr = raw_property[separator_index + 1 :].strip()
    if not raw_type_expr:
        msg = f"Alias {alias_name} contains empty type for property: {raw_name!r}"
        raise ValueError(msg)

    optional = raw_name.endswith("?")
    name = raw_name[:-1].strip() if optional else raw_name
    if not _IDENTIFIER_PATTERN.fullmatch(name):
        msg = f"Alias {alias_name} contains unsupported property name: {name!r}"
        raise ValueError(msg)

    return name, optional, raw_type_expr


def _split_top_level(text: str, *, delimiter: str) -> list[str]:
    if len(delimiter) != 1:
        msg = "Delimiter must be a single character"
        raise ValueError(msg)

    parts: list[str] = []
    start_index = 0
    state = _DepthState()

    for index, char in enumerate(text):
        state.consume_char(char)
        if char == delimiter and _is_top_level(state):
            parts.append(text[start_index:index])
            start_index = index + 1

    parts.append(text[start_index:])
    return parts


def _find_top_level(text: str, *, target: str) -> int:
    if len(target) != 1:
        msg = "Target must be a single character"
        raise ValueError(msg)

    state = _DepthState()

    for index, char in enumerate(text):
        state.consume_char(char)
        if char == target and _is_top_level(state):
            return index

    return -1


def _parse_string_literal(token: str) -> str | None:
    stripped_token = token.strip()
    if len(stripped_token) < _MIN_STRING_LITERAL_LENGTH:
        return None

    quote = stripped_token[0]
    if quote not in {'"', "'"} or stripped_token[-1] != quote:
        return None

    return stripped_token[1:-1]


def _starts_block_comment(source: str, index: int) -> bool:
    return source[index : index + 2] == "/*"


def _starts_line_comment(source: str, index: int) -> bool:
    return source[index : index + 2] == "//"


def _skip_block_comment(source: str, index: int, *, result: list[str]) -> int:
    index += 2
    while index < len(source):
        if source[index : index + 2] == "*/":
            return index + 2
        if source[index] == "\n":
            result.append("\n")
        index += 1
    return index


def _skip_line_comment(source: str, index: int) -> int:
    index += 2
    while index < len(source) and source[index] != "\n":
        index += 1
    return index


def _depth_delta(char: str, *, opening: str, closing: str) -> int:
    if char == opening:
        return 1
    if char == closing:
        return -1
    return 0


def _is_top_level(state: _DepthState) -> bool:
    return (
        state.in_string is None
        and state.brace_depth == 0
        and state.bracket_depth == 0
        and state.paren_depth == 0
    )



from __future__ import annotations

import os
import sys
from pathlib import Path
from textwrap import dedent


def create_fake_codex_executable(tmp_path: Path) -> Path:
    target = tmp_path / "fake_codex_target.py"
    target.write_text(_SCRIPT, encoding="utf-8")

    if os.name == "nt":
        executable = tmp_path / "codex.bat"
        executable.write_text(
            f'@echo off\r\n"{sys.executable}" "{target}" %*\r\n',
            encoding="utf-8",
        )
        return executable

    executable = tmp_path / "codex"
    executable.write_text(
        f'#!/bin/sh\nexec "{sys.executable}" "{target}" "$@"\n',
        encoding="utf-8",
    )
    executable.chmod(0o755)
    return executable


_SCRIPT = dedent(
    """\
    from __future__ import annotations

    import json
    import os
    import sys
    from pathlib import Path


    def parse_args(argv: list[str]) -> tuple[list[str], str | None, str | None]:
        images: list[str] = []
        output_schema: str | None = None
        resume_id: str | None = None
        index = 0
        while index < len(argv):
            current = argv[index]
            if current == "--image" and index + 1 < len(argv):
                images.append(argv[index + 1])
                index += 2
                continue
            if current == "--output-schema" and index + 1 < len(argv):
                output_schema = argv[index + 1]
                index += 2
                continue
            if current == "resume" and index + 1 < len(argv):
                resume_id = argv[index + 1]
                index += 2
                continue
            index += 1
        return images, output_schema, resume_id


    def emit(payload: dict[str, object]) -> None:
        sys.stdout.write(json.dumps(payload) + "\\n")
        sys.stdout.flush()


    argv = sys.argv[1:]
    stdin_text = sys.stdin.read()
    mode = os.environ.get("FAKE_CODEX_MODE", "lines")
    images, output_schema, resume_id = parse_args(argv)

    if mode == "lines":
        lines = json.loads(os.environ.get("FAKE_LINES_JSON", '["line-1", "line-2"]'))
        for line in lines:
            sys.stdout.write(str(line) + "\\n")
        sys.stdout.flush()
        raise SystemExit(0)

    if mode == "env_guard":
        marker_name = os.environ.get("FAKE_MARKER_NAME", "SHOULD_NOT_LEAK")
        leaked = marker_name in os.environ
        sys.stdout.write(("leaked" if leaked else "clean") + "\\n")
        sys.stdout.flush()
        raise SystemExit(0)

    if mode == "thread_success":
        thread_id = os.environ.get("FAKE_THREAD_ID", "thread-from-fake")
        responses = json.loads(os.environ.get("FAKE_RESPONSES_JSON", '["first", "second"]'))
        emit({"type": "thread.started", "thread_id": thread_id})
        emit({"type": "turn.started"})
        for index, text in enumerate(responses):
            emit(
                {
                    "type": "item.completed",
                    "item": {
                        "type": "agent_message",
                        "id": f"item-{index}",
                        "text": str(text),
                    },
                },
            )
        emit(
            {
                "type": "turn.completed",
                "usage": {
                    "input_tokens": 10,
                    "cached_input_tokens": 2,
                    "output_tokens": 5,
                },
            },
        )
        raise SystemExit(0)

    if mode == "thread_failed":
        message = os.environ.get("FAKE_FAILURE_MESSAGE", "failure from fake codex")
        emit({"type": "thread.started", "thread_id": "thread-failed"})
        emit({"type": "turn.started"})
        emit({"type": "turn.failed", "error": {"message": message}})
        raise SystemExit(0)

    if mode == "normalize_capture":
        payload = {
            "stdin": stdin_text,
            "images": images,
            "resume_id": resume_id,
            "argv": argv,
        }
        emit({"type": "thread.started", "thread_id": "thread-normalize"})
        emit(
            {
                "type": "item.completed",
                "item": {
                    "type": "agent_message",
                    "id": "capture",
                    "text": json.dumps(payload),
                },
            },
        )
        emit(
            {
                "type": "turn.completed",
                "usage": {
                    "input_tokens": 1,
                    "cached_input_tokens": 0,
                    "output_tokens": 1,
                },
            },
        )
        raise SystemExit(0)

    if mode == "schema_check":
        schema_exists = False
        schema_payload: object = None
        if output_schema is not None:
            schema_path = Path(output_schema)
            schema_exists = schema_path.exists()
            if schema_exists:
                schema_payload = json.loads(schema_path.read_text(encoding="utf-8"))
        emit({"type": "thread.started", "thread_id": "thread-schema"})
        emit(
            {
                "type": "item.completed",
                "item": {
                    "type": "agent_message",
                    "id": "schema",
                    "text": json.dumps(
                        {
                            "schema_path": output_schema,
                            "schema_exists": schema_exists,
                            "schema_payload": schema_payload,
                        },
                    ),
                },
            },
        )
        emit(
            {
                "type": "turn.completed",
                "usage": {
                    "input_tokens": 2,
                    "cached_input_tokens": 0,
                    "output_tokens": 2,
                },
            },
        )
        raise SystemExit(0)

    raise SystemExit(0)
    """,
)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal, TypeAlias

CommandExecutionStatus: TypeAlias = Literal["in_progress", "completed", "failed"]
PatchChangeKind: TypeAlias = Literal["add", "delete", "update"]
PatchApplyStatus: TypeAlias = Literal["completed", "failed"]
McpToolCallStatus: TypeAlias = Literal["in_progress", "completed", "failed"]


@dataclass(frozen=True, slots=True)
class CommandExecutionItem:
    """A command executed by the agent."""

    id: str
    """Stable identifier for this thread item."""

    command: str
    """Command line executed by the agent."""

    aggregated_output: str
    """Aggregated stdout and stderr captured while the command was running."""

    status: CommandExecutionStatus
    """Current status of the command execution."""

    exit_code: int | None = None
    """Exit code once the command exits; ``None`` while still running."""

    type: Literal["command_execution"] = field(default="command_execution", init=False)
    """Discriminator with value ``"command_execution"``."""


@dataclass(frozen=True, slots=True)
class FileUpdateChange:
    """A file update included in a patch operation."""

    path: str
    """Path to the changed file."""

    kind: PatchChangeKind
    """Kind of patch change (add, delete, or update)."""


@dataclass(frozen=True, slots=True)
class FileChangeItem:
    """A set of file changes by the agent.

    Emitted once the patch succeeds or fails.

    """

    id: str
    """Stable identifier for this thread item."""

    changes: list[FileUpdateChange]
    """Individual file changes that comprise the patch."""

    status: PatchApplyStatus
    """Whether the patch ultimately succeeded or failed."""

    type: Literal["file_change"] = field(default="file_change", init=False)
    """Discriminator with value ``"file_change"``."""


@dataclass(frozen=True, slots=True)
class McpToolCallResult:
    """Result payload returned by an MCP server."""

    content: list[object]
    """MCP content blocks returned for successful calls."""

    structured_content: object
    """Structured payload returned by the MCP server."""


@dataclass(frozen=True, slots=True)
class McpToolCallError:
    """Error payload returned for a failed MCP call."""

    message: str
    """Error message reported by the MCP server."""


@dataclass(frozen=True, slots=True)
class McpToolCallItem:
    """Represents a call to an MCP tool.

    The item starts when invocation is dispatched and completes when the MCP server reports
    success or failure.

    """

    id: str
    """Stable identifier for this thread item."""

    server: str
    """Name of the MCP server handling the request."""

    tool: str
    """Tool invoked on the MCP server."""

    arguments: object
    """Arguments forwarded to the tool invocation."""

    status: McpToolCallStatus
    """Current status of the tool invocation."""

    result: McpToolCallResult | None = None
    """Result payload for successful calls; ``None`` when not available."""

    error: McpToolCallError | None = None
    """Error payload for failed calls; ``None`` when not available."""

    type: Literal["mcp_tool_call"] = field(default="mcp_tool_call", init=False)
    """Discriminator with value ``"mcp_tool_call"``."""


@dataclass(frozen=True, slots=True)
class AgentMessageItem:
    """Response from the agent.

    The text can be natural language or JSON when structured output is requested.

    """

    id: str
    """Stable identifier for this thread item."""

    text: str
    """Agent response payload."""

    type: Literal["agent_message"] = field(default="agent_message", init=False)
    """Discriminator with value ``"agent_message"``."""


@dataclass(frozen=True, slots=True)
class ReasoningItem:
    """Agent reasoning summary."""

    id: str
    """Stable identifier for this thread item."""

    text: str
    """Reasoning summary text."""

    type: Literal["reasoning"] = field(default="reasoning", init=False)
    """Discriminator with value ``"reasoning"``."""


@dataclass(frozen=True, slots=True)
class WebSearchItem:
    """Captures a web search request.

    The item completes when results are returned to the agent.

    """

    id: str
    """Stable identifier for this thread item."""

    query: str
    """Search query text."""

    type: Literal["web_search"] = field(default="web_search", init=False)
    """Discriminator with value ``"web_search"``."""


@dataclass(frozen=True, slots=True)
class ErrorItem:
    """Describes a non-fatal error surfaced as an item."""

    id: str
    """Stable identifier for this thread item."""

    message: str
    """Error message."""

    type: Literal["error"] = field(default="error", init=False)
    """Discriminator with value ``"error"``."""


@dataclass(frozen=True, slots=True)
class TodoItem:
    """An item in the agent to-do list."""

    text: str
    """To-do item text."""

    completed: bool
    """Whether the item is completed."""


@dataclass(frozen=True, slots=True)
class TodoListItem:
    """Tracks the agent running to-do list.

    Starts when a plan is issued, updates as steps change, and completes when the turn ends.

    """

    id: str
    """Stable identifier for this thread item."""

    items: list[TodoItem]
    """Current to-do list items."""

    type: Literal["todo_list"] = field(default="todo_list", init=False)
    """Discriminator with value ``"todo_list"``."""


ThreadItem: TypeAlias = (
    AgentMessageItem
    | ReasoningItem
    | CommandExecutionItem
    | FileChangeItem
    | McpToolCallItem
    | WebSearchItem
    | TodoListItem
    | ErrorItem
)



from __future__ import annotations

import re

_EXPORT_BLOCK_PATTERN: re.Pattern[str] = re.compile(
    r"""
    export                     # export keyword
    \s+
    (?:type\s+)?               # optional 'type'
    \{                         # opening brace
    (?P<names>[^}]+)           # everything until closing brace
    \}                         # closing brace
    """,
    re.VERBOSE | re.MULTILINE,
)


def _parse_export_block(block: str) -> list[str]:
    results: list[str] = []

    for part in block.split(","):
        item: str = part.strip()
        if not item:
            continue

        # Handle aliasing: Foo as Bar -> export name is Bar
        if " as " in item:
            _, alias = item.split(" as ", 1)
            results.append(alias.strip())
        else:
            results.append(item)

    return results


def extract_exported_objects(ts_source: str) -> list[str]:
    """Extract all exported object names from TypeScript export blocks.

    This function looks for export statements of the form:
    export { Foo, Bar } from "./module";
    export type { Baz } from "./module";

    It captures the names of the exported objects (Foo, Bar, Baz) and returns them as a list.

    Returns:
        A list of exported object names.

    """
    exports: list[str] = []

    for match in _EXPORT_BLOCK_PATTERN.finditer(ts_source):
        names_block: str = match.group("names")
        exports.extend(_parse_export_block(names_block))

    return exports



from __future__ import annotations

from typing import TYPE_CHECKING, TypeAlias, TypedDict

if TYPE_CHECKING:
    from typing_extensions import NotRequired

# fmt: off
CodexConfigValue: TypeAlias = str | int | float | bool | list["CodexConfigValue"] | dict[str, "CodexConfigValue"]
CodexConfigObject: TypeAlias = dict[str, CodexConfigValue]
# fmt: on


class CodexOptions(TypedDict):
    """Options used to construct a Codex client."""

    codex_path_override: NotRequired[str]
    """Optional path to the Codex executable."""

    base_url: NotRequired[str]
    """Optional base URL used by Codex API calls."""

    api_key: NotRequired[str]
    """Optional API key used by Codex API calls."""

    config: NotRequired[CodexConfigObject]
    """Additional ``--config key=value`` overrides passed to Codex CLI.

    Provide a JSON-like object and the SDK will flatten dotted paths and serialize values as TOML
    literals for CLI compatibility.
    """

    env: NotRequired[dict[str, str]]
    """Environment variables passed to the Codex CLI process.

    When provided, the SDK does not inherit from the current process environment.
    """



from __future__ import annotations

from typing import TYPE_CHECKING

from acodex.exec import AsyncCodexExec, CodexExec
from acodex.thread import AsyncThread, Thread
from acodex.types.codex_options import CodexOptions
from acodex.types.thread_options import ThreadOptions

if TYPE_CHECKING:
    from typing_extensions import Unpack


class Codex:
    """Codex is the main class for interacting with the Codex agent.

    Use `start_thread()` to start a new thread or `resume_thread()` to resume a previously started
    thread.
    """

    def __init__(self, **options: Unpack[CodexOptions]) -> None:
        self._options = options
        self._exec = CodexExec(
            executable_path=self._options.get("codex_path_override"),
            env=self._options.get("env"),
            config_overrides=self._options.get("config"),
        )

    def start_thread(self, **thread_options: Unpack[ThreadOptions]) -> Thread:
        """Start a new conversation with an agent.

        Returns:
            A new thread instance.

        """
        return Thread(
            exec=self._exec,
            options=self._options,
            thread_options=thread_options,
        )

    def resume_thread(self, thread_id: str, **thread_options: Unpack[ThreadOptions]) -> Thread:
        """Resume a conversation with an agent using a thread ID.

        Threads are persisted in ``~/.codex/sessions``.

        Args:
            thread_id: Identifier of the thread to resume.

        Returns:
            A new thread instance.

        """
        return Thread(
            exec=self._exec,
            options=self._options,
            thread_options=thread_options,
            thread_id=thread_id,
        )


class AsyncCodex:
    """Codex is the main class for interacting with the Codex agent.

    Use `start_thread()` to start a new thread or `resume_thread()` to resume a previously started
    thread.
    """

    def __init__(self, **options: Unpack[CodexOptions]) -> None:
        self._options = options
        self._exec = AsyncCodexExec(
            executable_path=self._options.get("codex_path_override"),
            env=self._options.get("env"),
            config_overrides=self._options.get("config"),
        )

    def start_thread(self, **thread_options: Unpack[ThreadOptions]) -> AsyncThread:
        """Start a new conversation with an agent.

        Returns:
            A new thread instance.

        """
        return AsyncThread(
            exec=self._exec,
            options=self._options,
            thread_options=thread_options,
        )

    def resume_thread(self, thread_id: str, **options: Unpack[ThreadOptions]) -> AsyncThread:
        """Resume a conversation with an agent using a thread ID.

        Threads are persisted in ``~/.codex/sessions``.

        Args:
            thread_id: Identifier of the thread to resume.

        Returns:
            A new thread instance.

        """
        return AsyncThread(
            exec=self._exec,
            options=self._options,
            thread_options=options,
            thread_id=thread_id,
        )



from __future__ import annotations

import asyncio
import os
import queue
import re
import subprocess  # noqa: S404
import sys
import threading
from collections.abc import Coroutine
from pathlib import Path
from textwrap import dedent
from typing import Any, TypeVar, cast

import pytest

from acodex._internal.exec import CodexExecCommand
from acodex._internal.process_runner import (
    _ASYNC_CANCELLED,
    AsyncCodexProcessRunner,
    SyncCodexProcessRunner,
)
from acodex.exceptions import CodexCancelledError, CodexExecError

_T = TypeVar("_T")


def test_sync_stream_lines_yields_stdout_lines_without_newlines(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = SyncCodexProcessRunner(executable_path=str(executable))
    command = _build_command(mode="stdout_lines")

    lines = list(runner.stream_lines(command))

    assert lines == ["first line", "second line", "third line"]


def test_sync_nonzero_exit_raises_codex_exec_error_includes_stderr(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = SyncCodexProcessRunner(executable_path=str(executable))
    command = _build_command(mode="nonzero")

    with pytest.raises(CodexExecError, match=re.escape("Codex Exec exited with code 7")) as error:
        list(runner.stream_lines(command))

    assert error.value.stderr == "stderr boom\n"


def test_sync_nonzero_exit_before_stdout_closes_raises_codex_exec_error(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = SyncCodexProcessRunner(executable_path=str(executable))
    command = _build_command(mode="exit_before_stdout_closes")

    with pytest.raises(CodexExecError, match=re.escape("Codex Exec exited with code 7")) as error:
        list(runner.stream_lines(command))

    assert error.value.stderr == "stderr boom\n"


def test_sync_cancel_pre_set_raises_cancelled_without_spawning(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = SyncCodexProcessRunner(executable_path=str(executable))

    marker = tmp_path / "sync_pre_set_marker"
    signal = threading.Event()
    signal.set()

    command = _build_command(mode="stdout_lines", signal=signal, marker=marker)

    with pytest.raises(CodexCancelledError, match="Turn cancelled"):
        list(runner.stream_lines(command))

    assert not marker.exists()


def test_sync_cancel_mid_run_terminates_and_raises_cancelled(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = SyncCodexProcessRunner(executable_path=str(executable))

    signal = threading.Event()
    command = _build_command(mode="long_running", signal=signal)

    iterator = runner.stream_lines(command)

    assert next(iterator) == "ready"

    signal.set()
    with pytest.raises(CodexCancelledError, match="Turn cancelled"):
        next(iterator)


def test_sync_spawn_oserror_raises_codex_exec_error() -> None:
    runner = SyncCodexProcessRunner(executable_path="codex")
    command = _build_command(mode="stdout_lines")

    def raise_oserror(*_: object, **__: object) -> None:
        raise OSError("nope")

    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(subprocess, "Popen", raise_oserror)
        with pytest.raises(CodexExecError, match=re.escape("spawn failure: nope")):
            list(runner.stream_lines(command))


def test_sync_missing_stdio_streams_terminates_and_raises(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    runner = SyncCodexProcessRunner(executable_path="codex")
    command = _build_command(mode="stdout_lines")
    process = _MissingSyncStdioProcess()
    terminated: list[object] = []

    def fake_spawn(_: object) -> _MissingSyncStdioProcess:
        return process

    def fake_terminate(proc: object) -> None:
        terminated.append(proc)

    monkeypatch.setattr(runner, "_spawn_process", fake_spawn)
    monkeypatch.setattr(runner, "_terminate_process", fake_terminate)

    with pytest.raises(CodexExecError, match="spawn failure: missing stdio streams"):
        list(runner.stream_lines(command))

    assert terminated == [process]


def test_sync_stream_lines_handles_poll_timeouts_before_output(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    runner = SyncCodexProcessRunner(executable_path="codex")
    command = CodexExecCommand(argv=[], env={}, stdin="")
    stdout_queue: queue.Queue[Any] = queue.Queue()
    process = _PolledSyncProcess(return_code=0)
    stdout_thread = threading.Thread()

    monkeypatch.setattr("acodex._internal.process_runner._READ_POLL_INTERVAL_SECONDS", 0.001)

    lines = list(
        runner._iter_stdout_lines(
            process=cast("subprocess.Popen[str]", process),
            command=command,
            stdout_queue=cast("Any", stdout_queue),
            stdout_thread=stdout_thread,
        ),
    )

    assert lines == []


def test_sync_terminate_process_falls_back_to_kill_after_timeouts() -> None:
    process = _KillFallbackSyncProcess()

    SyncCodexProcessRunner._terminate_process(cast("subprocess.Popen[str]", process))

    assert process.kill_called


def test_async_stream_lines_yields_stdout_lines_without_newlines(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = AsyncCodexProcessRunner(executable_path=str(executable))
    command = _build_command(mode="stdout_lines")

    async def run() -> list[str]:
        return [line async for line in runner.stream_lines(command)]

    lines = _run_async(run())

    assert lines == ["first line", "second line", "third line"]


def test_async_nonzero_exit_raises_codex_exec_error_includes_stderr(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = AsyncCodexProcessRunner(executable_path=str(executable))
    command = _build_command(mode="nonzero")

    async def run() -> None:
        async for _ in runner.stream_lines(command):
            pass

    with pytest.raises(CodexExecError, match=re.escape("Codex Exec exited with code 7")) as error:
        _run_async(run())

    assert error.value.stderr == "stderr boom\n"


def test_async_nonzero_exit_before_stdout_closes_raises_codex_exec_error(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = AsyncCodexProcessRunner(executable_path=str(executable))
    command = _build_command(mode="exit_before_stdout_closes")

    async def run() -> None:
        async for _ in runner.stream_lines(command):
            pass

    with pytest.raises(CodexExecError, match=re.escape("Codex Exec exited with code 7")) as error:
        _run_async(asyncio.wait_for(run(), timeout=2.0))

    assert error.value.stderr == "stderr boom\n"


def test_async_cancel_pre_set_raises_cancelled_without_spawning(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = AsyncCodexProcessRunner(executable_path=str(executable))

    marker = tmp_path / "async_pre_set_marker"
    signal = asyncio.Event()
    signal.set()

    command = _build_command(mode="stdout_lines", signal=signal, marker=marker)

    async def run() -> None:
        async for _ in runner.stream_lines(command):
            pass

    with pytest.raises(CodexCancelledError, match="Turn cancelled"):
        _run_async(run())

    assert not marker.exists()


def test_async_cancel_mid_run_terminates_and_raises_cancelled(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = AsyncCodexProcessRunner(executable_path=str(executable))

    async def run() -> None:
        signal = asyncio.Event()
        command = _build_command(mode="long_running", signal=signal)
        iterator = runner.stream_lines(command)

        assert await anext(iterator) == "ready"

        signal.set()
        with pytest.raises(CodexCancelledError, match="Turn cancelled"):
            await anext(iterator)

    _run_async(run())


def test_async_spawn_oserror_raises_codex_exec_error() -> None:
    runner = AsyncCodexProcessRunner(executable_path="codex")
    command = _build_command(mode="stdout_lines")

    async def raise_oserror(*_: object, **__: object) -> asyncio.subprocess.Process:
        await asyncio.sleep(0)
        raise OSError("nope")

    async def run() -> None:
        async for _ in runner.stream_lines(command):
            pass

    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(asyncio, "create_subprocess_exec", raise_oserror)
        with pytest.raises(CodexExecError, match=re.escape("spawn failure: nope")):
            _run_async(run())


def test_async_require_stdio_raises_when_streams_missing() -> None:
    process = _MissingAsyncStdioProcess()

    with pytest.raises(CodexExecError, match="spawn failure: missing stdio streams"):
        AsyncCodexProcessRunner._require_stdio(cast("asyncio.subprocess.Process", process))


def test_async_stream_lines_missing_stdio_terminates_process(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    runner = AsyncCodexProcessRunner(executable_path="codex")
    command = _build_command(mode="stdout_lines")
    process = _MissingAsyncStdioProcess()
    terminated: list[object] = []

    async def fake_spawn(_: object) -> asyncio.subprocess.Process:
        await asyncio.sleep(0)
        return cast("asyncio.subprocess.Process", process)

    async def fake_terminate(proc: object) -> None:
        terminated.append(proc)
        await asyncio.sleep(0)

    async def run() -> None:
        async for _ in runner.stream_lines(command):
            pass

    monkeypatch.setattr(runner, "_spawn_process", fake_spawn)
    monkeypatch.setattr(runner, "_terminate_process", fake_terminate)

    with pytest.raises(CodexExecError, match="spawn failure: missing stdio streams"):
        _run_async(run())

    assert terminated == [process]


def test_async_stream_lines_create_task_failure_terminates_and_closes_stdin(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    runner = AsyncCodexProcessRunner(executable_path="codex")
    command = _build_command(mode="stdout_lines")
    stdin = _FakeAsyncStdin()
    process = _AsyncProcessWithStdio(stdin=stdin)
    terminated: list[object] = []
    real_create_task = asyncio.create_task

    async def fake_spawn(_: object) -> asyncio.subprocess.Process:
        await asyncio.sleep(0)
        return cast("asyncio.subprocess.Process", process)

    async def fake_terminate(proc: object) -> None:
        terminated.append(proc)
        await asyncio.sleep(0)

    def fake_create_task(
        coro: Coroutine[object, object, object],
        *,
        name: str | None = None,
    ) -> asyncio.Task[object]:
        if name == "acodex-async-stderr-reader":
            coro.close()
            raise RuntimeError("boom")
        return real_create_task(coro, name=name)

    async def run() -> None:
        async for _ in runner.stream_lines(command):
            pass

    monkeypatch.setattr(runner, "_spawn_process", fake_spawn)
    monkeypatch.setattr(runner, "_terminate_process", fake_terminate)
    monkeypatch.setattr(asyncio, "create_task", fake_create_task)

    with pytest.raises(RuntimeError, match="boom"):
        _run_async(run())

    assert terminated == [process]
    assert stdin.close_called


def test_async_cancel_during_readline_uses_async_event_wait_path(tmp_path: Path) -> None:
    executable = _create_fake_codex_executable(tmp_path)
    runner = AsyncCodexProcessRunner(executable_path=str(executable))

    async def run() -> None:
        signal = asyncio.Event()
        command = _build_command(mode="long_running", signal=signal)
        iterator = runner.stream_lines(command)

        assert await anext(iterator) == "ready"

        async def read_next() -> str:
            return await anext(iterator)

        next_task: asyncio.Task[str] = asyncio.create_task(read_next())
        await asyncio.sleep(0.01)
        assert not next_task.done()

        signal.set()
        with pytest.raises(CodexCancelledError, match="Turn cancelled"):
            await next_task

    _run_async(run())


def test_async_read_timeout_path_returns_timeout_then_eventually_yields(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    runner = AsyncCodexProcessRunner(executable_path="codex")
    command = CodexExecCommand(argv=[], env={}, stdin="", signal=threading.Event())
    stdout = _DelayedAsyncStdout()

    async def run() -> list[str]:
        return [
            line
            async for line in runner._iter_stdout_lines(
                command=command,
                stdout=cast("asyncio.StreamReader", stdout),
            )
        ]

    monkeypatch.setattr("acodex._internal.process_runner._READ_POLL_INTERVAL_SECONDS", 0.001)
    lines = _run_async(run())

    assert lines == ["eventual line"]


def test_async_cleanup_closes_stdin_and_cancels_stderr_task(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    runner = AsyncCodexProcessRunner(executable_path="codex")
    stdin = _FakeAsyncStdin()
    stderr_task: asyncio.Task[bytes] | None = None

    async def fake_terminate(_: object) -> None:
        await asyncio.sleep(0)

    async def run() -> None:
        nonlocal stderr_task
        stderr_task = asyncio.create_task(asyncio.sleep(10.0, result=b"stderr"))
        monkeypatch.setattr(runner, "_terminate_process", fake_terminate)
        await runner._cleanup(
            process=cast("asyncio.subprocess.Process", _FakeAsyncProcess()),
            stdin=cast("asyncio.StreamWriter", stdin),
            stderr_task=stderr_task,
        )

    _run_async(run())

    assert stdin.close_called
    assert stderr_task is not None
    assert stderr_task.cancelled()


def test_async_read_next_line_returns_cancelled_when_signal_task_finishes_first() -> None:
    runner = AsyncCodexProcessRunner(executable_path="codex")
    signal = asyncio.Event()
    signal.set()
    command = CodexExecCommand(argv=[], env={}, stdin="", signal=signal)
    stdout = _ImmediateAsyncStdout()

    result = _run_async(
        runner._read_next_line(command=command, stdout=cast("asyncio.StreamReader", stdout)),
    )

    assert result is _ASYNC_CANCELLED


def _build_command(
    *,
    mode: str,
    signal: threading.Event | asyncio.Event | None = None,
    marker: Path | None = None,
) -> CodexExecCommand:
    env = dict(os.environ)
    env["FAKE_CODEX_MODE"] = mode
    if marker is not None:
        env["FAKE_CODEX_MARKER"] = str(marker)

    return CodexExecCommand(
        argv=["exec", "--experimental-json"],
        env=env,
        stdin="ignored stdin",
        signal=signal,
    )


def _create_fake_codex_executable(tmp_path: Path) -> Path:
    target = tmp_path / "fake_codex_target.py"
    target.write_text(
        dedent(
            """\
            from __future__ import annotations

            import os
            import subprocess
            import sys
            import time
            from pathlib import Path

            marker = os.environ.get("FAKE_CODEX_MARKER")
            if marker:
                Path(marker).write_text("invoked", encoding="utf-8")

            mode = os.environ.get("FAKE_CODEX_MODE", "stdout_lines")
            _ = sys.stdin.read()

            if mode == "stdout_lines":
                sys.stdout.write("first line\\r\\nsecond line\\nthird line\\r\\n")
                sys.stdout.flush()
                raise SystemExit(0)

            if mode == "nonzero":
                sys.stderr.write("stderr boom\\n")
                sys.stderr.flush()
                raise SystemExit(7)

            if mode == "exit_before_stdout_closes":
                # Keep stdout alive briefly after parent exit to mirror TS parity race.
                subprocess.Popen(
                    [
                        sys.executable,
                        "-c",
                        "import sys, time; time.sleep(0.05); sys.stdout.write('late line\\\\n'); sys.stdout.flush()",
                    ],
                    stdin=subprocess.DEVNULL,
                )
                sys.stderr.write("stderr boom\\n")
                sys.stderr.flush()
                raise SystemExit(7)

            if mode == "long_running":
                sys.stdout.write("ready\\n")
                sys.stdout.flush()
                while True:
                    time.sleep(0.1)

            raise SystemExit(0)
            """,
        ),
        encoding="utf-8",
    )

    if os.name == "nt":
        executable = tmp_path / "codex.bat"
        executable.write_text(
            f'@echo off\r\n"{sys.executable}" "{target}" %*\r\n',
            encoding="utf-8",
        )
        return executable

    executable = tmp_path / "codex"
    executable.write_text(
        f'#!/bin/sh\nexec "{sys.executable}" "{target}" "$@"\n',
        encoding="utf-8",
    )
    executable.chmod(0o755)
    return executable


class _MissingSyncStdioProcess:
    stdin: None = None
    stdout: None = None
    stderr: None = None


class _PolledSyncProcess:
    def __init__(self, *, return_code: int) -> None:
        self._return_code = return_code

    def poll(self) -> int:
        return self._return_code


class _KillFallbackSyncProcess:
    def __init__(self) -> None:
        self.kill_called = False
        self.terminate_called = False

    def poll(self) -> int | None:
        if self.kill_called:
            return 0
        return None

    def terminate(self) -> None:
        self.terminate_called = True

    def wait(self, timeout: float | None = None) -> int:
        if not self.kill_called:
            raise subprocess.TimeoutExpired(cmd="codex", timeout=timeout or 1.0)
        return 0

    def kill(self) -> None:
        self.kill_called = True


class _MissingAsyncStdioProcess:
    stdin: None = None
    stdout: None = None
    stderr: None = None


class _FakeAsyncProcess:
    returncode = 0


class _AsyncProcessWithStdio:
    def __init__(self, *, stdin: _FakeAsyncStdin) -> None:
        self.stdin = stdin
        self.stdout = _UnusedAsyncStdout()
        self.stderr = _ReadableAsyncStderr()


class _FakeAsyncStdin:
    def __init__(self) -> None:
        self.close_called = False

    def is_closing(self) -> bool:
        return self.close_called

    def close(self) -> None:
        self.close_called = True

    async def wait_closed(self) -> None:
        _ = self.close_called


class _DelayedAsyncStdout:
    def __init__(self) -> None:
        self._read_calls = 0

    async def readline(self) -> bytes:
        self._read_calls += 1
        if self._read_calls == 1:
            await asyncio.sleep(0.01)
            return b"delayed line\n"
        if self._read_calls == 2:
            return b"eventual line\n"
        return b""


class _ImmediateAsyncStdout:
    def __init__(self) -> None:
        self._read_calls = 0

    async def readline(self) -> bytes:
        self._read_calls += 1
        return b"ready\n"


class _UnusedAsyncStdout:
    def __init__(self) -> None:
        self.read_calls = 0

    async def readline(self) -> bytes:
        self.read_calls += 1
        return b""


class _ReadableAsyncStderr:
    def __init__(self) -> None:
        self.read_calls = 0

    async def read(self) -> bytes:
        self.read_calls += 1
        return b"stderr"


def _run_async(awaitable: Coroutine[object, object, _T]) -> _T:
    return asyncio.run(awaitable)



from __future__ import annotations

import re
import threading
from itertools import pairwise
from typing import cast

import pytest

from acodex._internal.exec import (
    _INTERNAL_ORIGINATOR_ENV,
    _PYTHON_SDK_ORIGINATOR,
    CodexExecArgs,
    CodexExecCLICommandBuilder,
    build_exec_args,
)
from acodex.exceptions import CodexConfigError
from acodex.types.codex_options import CodexConfigObject
from acodex.types.input import UserInputLocalImage, UserInputText
from acodex.types.thread_options import ThreadOptions


def test_exec_cli_command_builder_dont_miss_args() -> None:
    builder = CodexExecCLICommandBuilder(args=CodexExecArgs(input="hello world"))

    command = builder.build_command()
    assert command.argv[0] == "exec"
    assert command.stdin == "hello world"

    exec_args = set(CodexExecArgs.__annotations__)
    assert exec_args - set(builder._seen_args) == set()


def test_exec_builder_allows_empty_input_string() -> None:
    builder = CodexExecCLICommandBuilder(args=CodexExecArgs(input=""))

    command = builder.build_command()

    assert not command.stdin
    assert command.argv[:2] == ["exec", "--experimental-json"]


def test_exec_builder_sets_base_url_and_api_key_env() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello", base_url="https://example.test", api_key="key-123"),
        env_overrides={},
    )

    command = builder.build_command()

    assert command.env["OPENAI_BASE_URL"] == "https://example.test"
    assert command.env["CODEX_API_KEY"] == "key-123"


def test_exec_builder_forwards_signal_object() -> None:
    signal = threading.Event()
    builder = CodexExecCLICommandBuilder(args=CodexExecArgs(input="hello", signal=signal))

    command = builder.build_command()

    assert command.signal is signal


def test_exec_builder_adds_exec_flags_model_sandbox_cd_add_dir_skip_git_schema() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(
            input="hello",
            model="gpt-5",
            sandbox_mode="workspace-write",
            working_directory="/workspace",
            additional_directories=["/extra/a", "/extra/b"],
            skip_git_repo_check=True,
            output_schema_file="/workspace/schema.json",
        ),
    )

    command = builder.build_command()

    assert _extract_flag_values(command.argv, "--model") == ["gpt-5"]
    assert _extract_flag_values(command.argv, "--sandbox") == ["workspace-write"]
    assert _extract_flag_values(command.argv, "--cd") == ["/workspace"]
    assert _extract_flag_values(command.argv, "--add-dir") == ["/extra/a", "/extra/b"]
    assert "--skip-git-repo-check" in command.argv
    assert _extract_flag_values(command.argv, "--output-schema") == ["/workspace/schema.json"]


def test_exec_builder_web_search_enabled_true_false_sets_config_override() -> None:
    enabled_builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello", web_search_enabled=True),
    )
    disabled_builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello", web_search_enabled=False),
    )

    enabled_command = enabled_builder.build_command()
    disabled_command = disabled_builder.build_command()

    assert _collect_config_values(enabled_command.argv, "web_search") == ['web_search="live"']
    assert _collect_config_values(disabled_command.argv, "web_search") == [
        'web_search="disabled"',
    ]


def test_exec_builder_places_resume_args_before_image_args() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(
            input="hello",
            thread_id="thread-1",
            images=["/images/a.png"],
        ),
    )

    command = builder.build_command()

    assert command.argv.index("resume") < command.argv.index("--image")


def test_exec_args_include_thead_args() -> None:
    thead_args = set(CodexExecArgs.__annotations__)
    thread_options_args = set(ThreadOptions.__annotations__)

    assert thread_options_args.issubset(thead_args)


def test_build_exec_args_includes_normalized_input_thread_metadata_and_turn_options() -> None:
    signal = threading.Event()
    args = build_exec_args(
        input=[
            UserInputText(text="hello"),
            UserInputLocalImage(path="/images/one.png"),
            UserInputText(text="world"),
        ],
        options={},
        thread_options={},
        thread_id="thread-123",
        turn_options={"signal": signal},
        output_schema_path="/schemas/schema.json",
    )

    assert args["input"] == "hello\n\nworld"
    assert args["images"] == ["/images/one.png"]
    assert args["thread_id"] == "thread-123"
    assert args["signal"] is signal
    assert args["output_schema_file"] == "/schemas/schema.json"


def test_build_exec_args_unpacks_thread_options_with_none_values() -> None:
    thread_options = cast(
        "ThreadOptions",
        {
            "model": None,
            "skip_git_repo_check": None,
            "network_access_enabled": None,
        },
    )

    args = build_exec_args(
        input="hello world",
        options={},
        thread_options=thread_options,
        thread_id=None,
        turn_options={},
        output_schema_path=None,
    )
    args_dict = cast("dict[str, object]", args)

    assert args_dict["model"] is None
    assert args_dict["skip_git_repo_check"] is None
    assert args_dict["network_access_enabled"] is None


def test_build_exec_args_includes_base_url_and_api_key_from_codex_options() -> None:
    args = build_exec_args(
        input="hello world",
        options={"base_url": "https://example.test", "api_key": "key-123"},
        thread_options={},
        thread_id=None,
        turn_options={},
        output_schema_path=None,
    )

    assert args["base_url"] == "https://example.test"
    assert args["api_key"] == "key-123"


def test_exec_builder_serializes_config_overrides() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello world"),
        config_overrides={
            "retry_budget": 3,
            "approval_policy": "never",
            "sandbox_workspace_write": {"network_access": True},
            "tool_rules": {"allow": ["git status", "git diff"]},
        },
    )

    command = builder.build_command()

    assert _collect_all_config_values(command.argv) == [
        "retry_budget=3",
        'approval_policy="never"',
        "sandbox_workspace_write.network_access=true",
        'tool_rules.allow=["git status", "git diff"]',
    ]


def test_exec_builder_serializes_numeric_overrides_with_js_number_parity() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello world"),
        config_overrides={
            "small": 1e-7,
            "threshold": 1e-6,
        },
    )

    command = builder.build_command()

    assert _collect_all_config_values(command.argv) == [
        "small=1e-7",
        "threshold=0.000001",
    ]


def test_exec_builder_serializes_empty_nested_object_override() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello world"),
        config_overrides={"a": {"b": {}}},
    )

    command = builder.build_command()

    assert _collect_all_config_values(command.argv) == ["a.b={}"]


def test_exec_builder_skips_empty_top_level_config_overrides() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello world"),
        config_overrides={},
    )

    command = builder.build_command()

    assert "--config" not in command.argv


@pytest.mark.parametrize(
    ("config_overrides", "error"),
    [
        (cast("CodexConfigObject", 1), "Codex config overrides must be a plain object"),
        (
            cast("CodexConfigObject", {"": 1}),
            "Codex config override keys must be non-empty strings",
        ),
        (
            cast("CodexConfigObject", {"numbers": [float("inf")]}),
            "Codex config override at numbers[0] must be a finite number",
        ),
        (
            cast("CodexConfigObject", {"root": None}),
            "Codex config override at root cannot be null",
        ),
        (
            cast("CodexConfigObject", {"root": (1, 2)}),
            "Unsupported Codex config override value at root: tuple",
        ),
    ],
)
def test_exec_builder_raises_for_invalid_config_overrides(
    config_overrides: CodexConfigObject,
    error: str,
) -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello world"),
        config_overrides=config_overrides,
    )

    with pytest.raises(CodexConfigError, match=re.escape(error)):
        builder.build_command()


def test_exec_builder_keeps_config_override_precedence_order() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello world", approval_policy="on-request"),
        config_overrides={"approval_policy": "never"},
    )

    command = builder.build_command()
    approval_policy_overrides = _collect_config_values(command.argv, "approval_policy")

    assert approval_policy_overrides == [
        'approval_policy="never"',
        'approval_policy="on-request"',
    ]


def test_exec_builder_serializes_thread_config_flags_as_config_values() -> None:
    builder = CodexExecCLICommandBuilder(
        args=cast(
            "CodexExecArgs",
            {
                "input": "hello world",
                "model_reasoning_effort": 'high"value',
                "network_access_enabled": True,
                "web_search_mode": 'live"value',
                "approval_policy": 'never"value',
            },
        ),
    )

    command = builder.build_command()

    assert _collect_all_config_values(command.argv) == [
        'model_reasoning_effort="high\\"value"',
        "sandbox_workspace_write.network_access=true",
        'web_search="live\\"value"',
        'approval_policy="never\\"value"',
    ]


def test_exec_builder_does_not_inherit_process_env_when_empty_env_overrides_provided(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("CODEX_ENV_SHOULD_NOT_LEAK", "leak")

    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello world"),
        env_overrides={},
    )

    command = builder.build_command()

    assert command.env == {
        _INTERNAL_ORIGINATOR_ENV: _PYTHON_SDK_ORIGINATOR,
    }


def test_exec_builder_keeps_existing_internal_originator_env_value() -> None:
    builder = CodexExecCLICommandBuilder(
        args=CodexExecArgs(input="hello world"),
        env_overrides={_INTERNAL_ORIGINATOR_ENV: "custom-originator"},
    )

    command = builder.build_command()

    assert command.env[_INTERNAL_ORIGINATOR_ENV] == "custom-originator"


def _collect_all_config_values(argv: list[str]) -> list[str]:
    return [value for key, value in pairwise(argv) if key == "--config"]


def _extract_flag_values(argv: list[str], flag: str) -> list[str]:
    values: list[str] = []
    index = 0
    while index < len(argv):
        if argv[index] == flag and index + 1 < len(argv):
            values.append(argv[index + 1])
            index += 2
            continue
        index += 1
    return values


def _collect_config_values(argv: list[str], key: str) -> list[str]:
    prefix = f"{key}="
    return [value for value in _collect_all_config_values(argv) if value.startswith(prefix)]



from __future__ import annotations

import asyncio
import json
import re
from pathlib import Path
from typing import cast

import pytest

from acodex.exceptions import CodexThreadRunError
from acodex.exec import AsyncCodexExec
from acodex.thread import AsyncThread, _aclose_if_possible
from acodex.types.events import ItemCompletedEvent, ThreadStartedEvent
from acodex.types.input import Input, UserInputLocalImage, UserInputText
from acodex.types.items import AgentMessageItem
from acodex.types.turn_options import OutputSchemaInput
from tests.unit.fake_codex_executable import create_fake_codex_executable


def test_async_thread_run_streamed_yields_events_and_sets_thread_id(tmp_path: Path) -> None:
    thread = _build_thread(
        tmp_path,
        env={"FAKE_CODEX_MODE": "thread_success", "FAKE_THREAD_ID": "async-thread-123"},
    )

    async def run() -> list[object]:
        streamed = await thread.run_streamed("hello")
        return [event async for event in streamed.events]

    events = asyncio.run(run())

    assert isinstance(events[0], ThreadStartedEvent)
    assert thread.id == "async-thread-123"


def test_async_thread_run_returns_completed_turn_with_final_response_and_usage(
    tmp_path: Path,
) -> None:
    thread = _build_thread(
        tmp_path,
        env={
            "FAKE_CODEX_MODE": "thread_success",
            "FAKE_RESPONSES_JSON": json.dumps(["draft", "async final"]),
        },
    )

    async def run() -> tuple[str, int | None]:
        turn = await thread.run("hello")
        output_tokens = turn.usage.output_tokens if turn.usage is not None else None
        return turn.final_response, output_tokens

    final_response, output_tokens = asyncio.run(run())

    assert final_response == "async final"
    assert output_tokens == 5


def test_async_thread_run_raises_codex_thread_run_error_on_turn_failed(tmp_path: Path) -> None:
    message = "async CLI failure"
    thread = _build_thread(
        tmp_path,
        env={"FAKE_CODEX_MODE": "thread_failed", "FAKE_FAILURE_MESSAGE": message},
    )

    async def run() -> None:
        await thread.run("hello")

    with pytest.raises(CodexThreadRunError, match=re.escape(message)):
        asyncio.run(run())


def test_async_thread_input_normalization_joins_text_and_passes_images_in_order(
    tmp_path: Path,
) -> None:
    thread = _build_thread(
        tmp_path,
        env={"FAKE_CODEX_MODE": "normalize_capture"},
        thread_id="async-resume-1",
    )

    async def run() -> dict[str, object]:
        turn = await thread.run(
            [
                UserInputText(text="first paragraph"),
                UserInputLocalImage(path="/images/a.png"),
                UserInputText(text="second paragraph"),
                UserInputLocalImage(path="/images/b.png"),
            ],
        )
        return cast("dict[str, object]", json.loads(turn.final_response))

    payload = asyncio.run(run())

    assert payload["stdin"] == "first paragraph\n\nsecond paragraph"
    assert payload["images"] == ["/images/a.png", "/images/b.png"]
    assert payload["resume_id"] == "async-resume-1"
    assert _extract_flag_values(cast("list[str]", payload["argv"]), "--image") == [
        "/images/a.png",
        "/images/b.png",
    ]


def test_async_thread_output_schema_file_lifecycle_exhausted_and_closed_stream(
    tmp_path: Path,
) -> None:
    schema: OutputSchemaInput = {"type": "object", "properties": {"ok": {"type": "boolean"}}}

    exhausted_thread = _build_thread(tmp_path, env={"FAKE_CODEX_MODE": "schema_check"})
    closed_thread = _build_thread(tmp_path, env={"FAKE_CODEX_MODE": "schema_check"})

    async def run_exhausted() -> dict[str, object]:
        turn = await exhausted_thread.run("hello", output_schema=schema)
        return cast("dict[str, object]", json.loads(turn.final_response))

    exhausted_payload = asyncio.run(run_exhausted())
    exhausted_path = Path(cast("str", exhausted_payload["schema_path"]))
    assert exhausted_payload["schema_exists"] is True
    assert exhausted_payload["schema_payload"] == schema
    assert not exhausted_path.exists()
    assert not exhausted_path.parent.exists()

    async def run_closed() -> tuple[Path, bool]:
        streamed = await closed_thread.run_streamed("hello", output_schema=schema)
        events = streamed.events
        _ = await anext(events)
        payload_event = await anext(events)
        assert isinstance(payload_event, ItemCompletedEvent)
        assert isinstance(payload_event.item, AgentMessageItem)
        payload = cast("dict[str, object]", json.loads(payload_event.item.text))
        schema_path = Path(cast("str", payload["schema_path"]))
        before_close_exists = bool(payload["schema_exists"])
        close_method = getattr(events, "aclose", None)
        assert close_method is not None
        await close_method()
        return schema_path, before_close_exists

    closed_path, before_close_exists = asyncio.run(run_closed())
    assert before_close_exists
    assert not closed_path.exists()
    assert not closed_path.parent.exists()


def test_async_thread_run_streamed_skips_unknown_events(tmp_path: Path) -> None:
    thread = _build_thread(
        tmp_path,
        env={
            "FAKE_CODEX_MODE": "lines",
            "FAKE_LINES_JSON": json.dumps(
                [
                    json.dumps({"type": "future.event", "payload": {"x": 1}}),
                    json.dumps({"type": "thread.started", "thread_id": "thread-async-known"}),
                ],
            ),
        },
    )

    async def run() -> list[object]:
        streamed = await thread.run_streamed("hello")
        return [event async for event in streamed.events]

    events = asyncio.run(run())

    assert len(events) == 1
    assert isinstance(events[0], ThreadStartedEvent)
    assert thread.id == "thread-async-known"


def test_async_thread_run_streamed_cleanup_when_build_exec_args_raises(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    thread = _build_thread(tmp_path, env={"FAKE_CODEX_MODE": "thread_success"})
    schema: OutputSchemaInput = {"type": "object", "properties": {"ok": {"type": "boolean"}}}
    forced_temp_dir = tmp_path / "forced-async-schema-dir"

    def fake_mkdtemp(*_: object, **__: object) -> str:
        forced_temp_dir.mkdir(parents=True, exist_ok=True)
        return str(forced_temp_dir)

    monkeypatch.setattr("acodex._internal.output_schema_file.tempfile.mkdtemp", fake_mkdtemp)

    async def run() -> None:
        streamed = await thread.run_streamed(cast("Input", {"bad": 1}), output_schema=schema)
        with pytest.raises(TypeError, match=r"input must be str or list\[UserInput\], got dict"):
            await anext(streamed.events)

    asyncio.run(run())

    assert not forced_temp_dir.exists()


def test_aclose_if_possible_is_noop_without_aclose() -> None:
    asyncio.run(_aclose_if_possible(object()))


def _build_thread(
    tmp_path: Path,
    *,
    env: dict[str, str],
    thread_id: str | None = None,
) -> AsyncThread:
    executable = create_fake_codex_executable(tmp_path)
    exec_client = AsyncCodexExec(executable_path=str(executable), env=env)
    return AsyncThread(exec=exec_client, options={}, thread_options={}, thread_id=thread_id)


def _extract_flag_values(argv: list[str], flag: str) -> list[str]:
    values: list[str] = []
    index = 0
    while index < len(argv):
        if argv[index] == flag and index + 1 < len(argv):
            values.append(argv[index + 1])
            index += 2
            continue
        index += 1
    return values



from __future__ import annotations

from typing import cast

import pytest

from acodex._internal.config import (
    _flatten_config_overrides,
    _normalize_exponent,
    serialize_config_overrides,
    to_config_value,
)
from acodex.exceptions import CodexConfigError
from acodex.types.codex_options import CodexConfigObject, CodexConfigValue


def _split_override(override: str) -> tuple[str, str]:
    path, _, value = override.partition("=")
    return path, value


def test_to_config_value_renders_string_like_json_stringify() -> None:
    assert to_config_value("hello", "value") == '"hello"'
    assert to_config_value("привет", "value") == '"привет"'


def test_to_config_value_renders_numbers_and_booleans() -> None:
    true_value = True
    false_value = False

    assert to_config_value(42, "value") == "42"
    assert to_config_value(1.5, "value") == "1.5"
    assert to_config_value(1.0, "value") == "1"
    assert to_config_value(true_value, "value") == "true"
    assert to_config_value(false_value, "value") == "false"


def test_to_config_value_renders_numbers_with_js_stringification_rules() -> None:
    assert to_config_value(-0.0, "value") == "0"
    assert to_config_value(1e-7, "value") == "1e-7"
    assert to_config_value(1e-6, "value") == "0.000001"
    assert to_config_value(1e16, "value") == "10000000000000000"
    assert to_config_value(1e21, "value") == "1e+21"


def test_to_config_value_renders_int_with_js_number_parity() -> None:
    assert (
        to_config_value(cast("CodexConfigValue", 9007199254740993), "value") == "9007199254740992"
    )


def test_to_config_value_exponential_mantissa_with_decimal_executes_normalization() -> None:
    assert to_config_value(1.23e21, "value") == "1.23e+21"


def test_normalize_exponent_accepts_missing_sign() -> None:
    assert _normalize_exponent("21") == "+21"


def test_to_config_value_renders_nested_array_and_object() -> None:
    value: CodexConfigValue = {
        "plain": "value",
        "needs.dot": {"nested-key": [1, True, "x"]},
    }
    assert (
        to_config_value(value, "value")
        == '{plain = "value", "needs.dot" = {nested-key = [1, true, "x"]}}'
    )


def test_config_overrides_nested_dict_flattens_to_dotted_key() -> None:
    config: CodexConfigObject = {"a": {"b": 1}}
    flattened: list[str] = []

    _flatten_config_overrides(config, "", overrides=flattened)
    serialized = serialize_config_overrides(config)

    assert flattened == ["a.b=1"]
    assert serialized == ["a.b=1"]
    path, value = _split_override(flattened[0])
    assert path == "a.b"
    assert value == "1"
    assert "a.b=1" in "\n".join(serialized).replace(" ", "")


def test_config_overrides_empty_root_dict_emits_no_overrides() -> None:
    empty: CodexConfigObject = {}
    flattened: list[str] = []

    _flatten_config_overrides(empty, "", overrides=flattened)
    serialized = serialize_config_overrides(empty)

    assert flattened == []
    assert serialized == []
    assert not "\n".join(serialized)


def test_config_overrides_empty_nested_dict_emits_empty_mapping_override() -> None:
    empty_mapping: dict[str, CodexConfigValue] = {}
    config: CodexConfigObject = {"a": empty_mapping}
    flattened: list[str] = []

    _flatten_config_overrides(config, "", overrides=flattened)
    serialized = serialize_config_overrides(config)

    assert flattened == ["a={}"]
    assert serialized == ["a={}"]
    path, value = _split_override(flattened[0])
    assert path == "a"
    assert value == "{}"
    assert "a={}" in "\n".join(serialized).replace(" ", "")


def test_to_config_value_rejects_non_finite_number() -> None:
    with pytest.raises(
        CodexConfigError,
        match="Codex config override at value must be a finite number",
    ):
        to_config_value(float("inf"), "value")


def test_to_config_value_rejects_overflow_int_as_non_finite_number() -> None:
    with pytest.raises(
        CodexConfigError,
        match="Codex config override at value must be a finite number",
    ):
        to_config_value(cast("CodexConfigValue", 10**400), "value")


def test_to_config_value_rejects_null_value_with_path() -> None:
    value = cast("CodexConfigValue", {"root": None})
    with pytest.raises(
        CodexConfigError,
        match=r"Codex config override at value\.root cannot be null",
    ):
        to_config_value(value, "value")


def test_to_config_value_rejects_empty_or_non_string_keys() -> None:
    with pytest.raises(
        CodexConfigError,
        match="Codex config override keys must be non-empty strings",
    ):
        to_config_value({"": 1}, "value")

    with pytest.raises(
        CodexConfigError,
        match="Codex config override keys must be non-empty strings",
    ):
        to_config_value(cast("CodexConfigValue", {1: "x"}), "value")


def test_to_config_value_rejects_unsupported_value_type() -> None:
    value = cast("CodexConfigValue", (1, 2))
    with pytest.raises(
        CodexConfigError,
        match=r"Unsupported Codex config override value at value: tuple",
    ):
        to_config_value(value, "value")



from __future__ import annotations

from acodex import __all__ as acodex_all
from tests.compatibility.vendor_ts_sdk import VENDOR_TS_SDK_SRC
from tools.compatibility.get_ts_exports import extract_exported_objects


def test_same_exports() -> None:
    ts_sdk_root = VENDOR_TS_SDK_SRC / "index.ts"

    exported_from_ts = set(extract_exported_objects(ts_sdk_root.read_text()))
    exported_from_python = set(acodex_all)

    assert exported_from_ts - exported_from_python == set(), (
        "Some exports are missing from Python SDK"
    )

    assert exported_from_python - exported_from_ts == {
        "AsyncCodex",
        "AsyncThread",
    }, "Some exports are missing from TypeScript SDK"



from __future__ import annotations

import asyncio
import threading
from typing import TYPE_CHECKING, TypeAlias, TypedDict

if TYPE_CHECKING:
    from typing_extensions import NotRequired

JsonObject: TypeAlias = dict[str, "JsonValue"]
JsonValue: TypeAlias = str | int | float | bool | list["JsonValue"] | JsonObject | None
TurnSignal: TypeAlias = threading.Event | asyncio.Event
OutputSchemaInput: TypeAlias = JsonObject


class TurnOptions(TypedDict):
    """Options for a single turn.

    Set the event (`event.set()`) to request cancellation.
    Use `threading.Event` for synchronous flows and `asyncio.Event` for asynchronous flows.
    """

    output_schema: NotRequired[OutputSchemaInput]
    signal: NotRequired[TurnSignal]



from __future__ import annotations

import argparse
import json
import os
import sys
import urllib.error
import urllib.request
from typing import Final

_REPO: Final[str] = "openai/codex"
_API_ROOT: Final[str] = "https://api.github.com"
_USER_AGENT: Final[str] = "acodex-vendor-bot"


class _GitHubApiError(RuntimeError):
    pass


def _github_get_json(*, url: str) -> object:
    headers: dict[str, str] = {
        "Accept": "application/vnd.github+json",
        "User-Agent": _USER_AGENT,
    }

    token = os.environ.get("GITHUB_TOKEN")
    if token:
        headers["Authorization"] = f"Bearer {token}"

    request = urllib.request.Request(url, headers=headers)  # noqa: S310
    try:
        with urllib.request.urlopen(request, timeout=30) as response:  # noqa: S310
            body = response.read().decode("utf-8")
    except urllib.error.HTTPError as exc:  # pragma: no cover
        raise _GitHubApiError(f"GitHub API error: {exc.code} {exc.reason} for {url}") from exc
    except urllib.error.URLError as exc:  # pragma: no cover
        raise _GitHubApiError(f"GitHub API request failed for {url}: {exc.reason}") from exc

    data: object = json.loads(body)
    return data


def _require_dict(value: object, *, context: str) -> dict[str, object]:
    if not isinstance(value, dict):
        raise _GitHubApiError(f"Expected object for {context}, got {type(value).__name__}")
    return value


def _require_list(value: object, *, context: str) -> list[object]:
    if not isinstance(value, list):
        raise _GitHubApiError(f"Expected array for {context}, got {type(value).__name__}")
    return value


def _require_str(obj: dict[str, object], key: str, *, context: str) -> str:
    value = obj.get(key)
    if not isinstance(value, str) or not value:
        raise _GitHubApiError(f"Missing/invalid {context}.{key}")
    return value


def _require_int(obj: dict[str, object], key: str, *, context: str) -> int:
    value = obj.get(key)
    if not isinstance(value, int) or value <= 0:
        raise _GitHubApiError(f"Missing/invalid {context}.{key}")
    return value


def _require_bool(obj: dict[str, object], key: str, *, context: str) -> bool:
    value = obj.get(key)
    if not isinstance(value, bool):
        raise _GitHubApiError(f"Missing/invalid {context}.{key}")
    return value


def latest_stable_release() -> dict[str, object]:
    page = 1
    while True:
        url = f"{_API_ROOT}/repos/{_REPO}/releases?per_page=100&page={page}"
        payload = _github_get_json(url=url)
        releases = _require_list(payload, context="releases")
        if not releases:
            break

        for idx, raw_release in enumerate(releases):
            context = f"releases[{page}:{idx}]"
            release = _require_dict(raw_release, context=context)
            if _require_bool(release, "draft", context=context):
                continue
            if _require_bool(release, "prerelease", context=context):
                continue

            return {
                "release_id": _require_int(release, "id", context=context),
                "release_tag": _require_str(release, "tag_name", context=context),
                "release_published_at": _require_str(release, "published_at", context=context),
                "release_url": _require_str(release, "html_url", context=context),
            }

        page += 1

    raise _GitHubApiError("No stable GitHub Releases found for openai/codex.")


def _parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Print the latest stable GitHub Release for openai/codex.",
    )
    parser.add_argument(
        "--field",
        choices=("release_id", "release_tag", "release_published_at", "release_url"),
        help="Print only a single field (plain text).",
    )
    return parser.parse_args(argv)


def main(argv: list[str]) -> int:
    args = _parse_args(argv)
    info = latest_stable_release()

    field: str | None = args.field
    if field is not None:
        value = info[field]
        sys.stdout.write(f"{value}\n")
        return 0

    sys.stdout.write(json.dumps(info, indent=2, sort_keys=True))
    sys.stdout.write("\n")
    return 0


if __name__ == "__main__":
    raise SystemExit(main(sys.argv[1:]))



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal, TypeAlias


@dataclass(frozen=True, slots=True)
class UserInputText:
    """A text input to send to the agent."""

    text: str
    type: Literal["text"] = field(default="text", init=False)


@dataclass(frozen=True, slots=True)
class UserInputLocalImage:
    """A local image input to send to the agent."""

    path: str
    type: Literal["local_image"] = field(default="local_image", init=False)


UserInput: TypeAlias = UserInputText | UserInputLocalImage
Input: TypeAlias = str | list[UserInput]



from __future__ import annotations

from pathlib import Path
from shutil import which

from acodex.exceptions import CodexExecutableNotFoundError


def find_codex_path() -> str:
    """Resolve the Codex CLI executable path from PATH.

    Returns:
        The discovered executable path as a string.

    Raises:
        CodexExecutableNotFoundError: If ``codex`` is not discoverable on ``PATH``.

    """
    # NOTE: The Windows Python<3.12 deprecation is for PathLike `cmd` values; we pass a plain
    # string literal ("codex"), so it does not apply here.
    found = which("codex")
    if found is None:
        raise CodexExecutableNotFoundError(
            'Could not locate "codex" on PATH. Install the Codex CLI or pass `codex_path_override`.',
        )

    return str(Path(found))



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Final, TypedDict

from acodex._internal.config import serialize_config_overrides, to_config_value
from acodex._internal.thread_core import normalize_input
from acodex.types.codex_options import CodexConfigObject, CodexConfigValue, CodexOptions
from acodex.types.input import Input
from acodex.types.thread_options import (
    ApprovalMode,
    ModelReasoningEffort,
    SandboxMode,
    ThreadOptions,
    WebSearchMode,
)
from acodex.types.turn_options import TurnOptions, TurnSignal

if TYPE_CHECKING:
    from typing_extensions import NotRequired


_INTERNAL_ORIGINATOR_ENV: Final[str] = "CODEX_INTERNAL_ORIGINATOR_OVERRIDE"
_PYTHON_SDK_ORIGINATOR: Final[str] = "codex_sdk_py"


class CodexExecArgs(TypedDict):
    """Arguments for the exec layer.

    Set `signal` via `event.set()` to request cancellation. Use `threading.Event` in synchronous
    flows and `asyncio.Event` in asynchronous flows.
    """

    input: str
    base_url: NotRequired[str]
    api_key: NotRequired[str]
    thread_id: NotRequired[str | None]
    images: NotRequired[list[str]]
    model: NotRequired[str]
    sandbox_mode: NotRequired[SandboxMode]
    working_directory: NotRequired[str]
    additional_directories: NotRequired[list[str]]
    skip_git_repo_check: NotRequired[bool]
    output_schema_file: NotRequired[str]
    model_reasoning_effort: NotRequired[ModelReasoningEffort]
    signal: NotRequired[TurnSignal]
    network_access_enabled: NotRequired[bool]
    web_search_mode: NotRequired[WebSearchMode]
    web_search_enabled: NotRequired[bool]
    approval_policy: NotRequired[ApprovalMode]


@dataclass
class CodexExecCommand:
    """Structured representation of a Codex CLI command."""

    argv: list[str]
    env: dict[str, str]
    stdin: str
    signal: TurnSignal | None = None


class CodexExecCLICommandBuilder:
    def __init__(
        self,
        *,
        args: CodexExecArgs,
        config_overrides: CodexConfigObject | None = None,
        env_overrides: dict[str, str] | None = None,
    ) -> None:
        self._args = args
        self._config_overrides = config_overrides
        self._env_overrides = env_overrides

        self._command = CodexExecCommand(argv=[], env={}, stdin="")
        self._seen_args: set[str] = set()

    def build_command(self) -> CodexExecCommand:
        self._add_input()
        self._add_env()
        self._add_signal()

        self._add_initial_command()
        self._add_config_overrides()
        self._add_model()
        self._add_sandbox_mode()
        self._add_working_directory()
        self._add_additional_directories()
        self._add_skip_git_repo_check()
        self._add_output_schema_file()
        self._add_model_reasoning_effort()
        self._add_network_access_enabled()
        self._add_web_search_mode()
        self._add_approval_policy()
        self._add_thread_id()
        self._add_images()

        return self._command

    def _add_input(self) -> None:
        self._seen_args.add("input")

        stdin = self._args.get("input")
        if not stdin:
            return

        self._command.stdin = stdin

    def _add_env(self) -> None:
        env = dict(self._env_overrides) if self._env_overrides is not None else dict(os.environ)

        self._seen_args.add("base_url")
        base_url = self._args.get("base_url")
        if base_url:
            env["OPENAI_BASE_URL"] = base_url

        self._seen_args.add("api_key")
        api_key = self._args.get("api_key")
        if api_key:
            env["CODEX_API_KEY"] = api_key

        if _INTERNAL_ORIGINATOR_ENV not in env:
            env[_INTERNAL_ORIGINATOR_ENV] = _PYTHON_SDK_ORIGINATOR

        self._command.env = env

    def _add_signal(self) -> None:
        self._seen_args.add("signal")

        signal = self._args.get("signal")
        if not signal:
            return

        self._command.signal = signal

    def _add_initial_command(self) -> None:
        self._command.argv.append("exec")
        self._command.argv.append("--experimental-json")

    def _add_config_overrides(self) -> None:
        if self._config_overrides is None:
            return

        for override in serialize_config_overrides(self._config_overrides):
            self._command.argv.append("--config")
            self._command.argv.append(override)

    def _add_model(self) -> None:
        self._seen_args.add("model")

        model = self._args.get("model")
        if not model:
            return

        self._command.argv.append("--model")
        self._command.argv.append(str(model))

    def _add_sandbox_mode(self) -> None:
        self._seen_args.add("sandbox_mode")

        sandbox_mode = self._args.get("sandbox_mode")
        if not sandbox_mode:
            return

        self._command.argv.append("--sandbox")
        self._command.argv.append(str(sandbox_mode))

    def _add_working_directory(self) -> None:
        self._seen_args.add("working_directory")

        working_directory = self._args.get("working_directory")
        if not working_directory:
            return

        self._command.argv.append("--cd")
        self._command.argv.append(str(working_directory))

    def _add_additional_directories(self) -> None:
        self._seen_args.add("additional_directories")

        additional_directories = self._args.get("additional_directories")
        if not additional_directories:
            return

        for directory in additional_directories:
            self._command.argv.append("--add-dir")
            self._command.argv.append(str(directory))

    def _add_skip_git_repo_check(self) -> None:
        self._seen_args.add("skip_git_repo_check")

        skip_git_repo_check = self._args.get("skip_git_repo_check")
        if not skip_git_repo_check:
            return

        self._command.argv.append("--skip-git-repo-check")

    def _add_output_schema_file(self) -> None:
        self._seen_args.add("output_schema_file")

        output_schema_file = self._args.get("output_schema_file")
        if not output_schema_file:
            return

        self._command.argv.append("--output-schema")
        self._command.argv.append(str(output_schema_file))

    def _add_model_reasoning_effort(self) -> None:
        self._seen_args.add("model_reasoning_effort")

        model_reasoning_effort = self._args.get("model_reasoning_effort")
        if not model_reasoning_effort:
            return

        self._append_config_override("model_reasoning_effort", model_reasoning_effort)

    def _add_network_access_enabled(self) -> None:
        self._seen_args.add("network_access_enabled")

        network_access_enabled = self._args.get("network_access_enabled")
        if network_access_enabled is None:
            return

        self._append_config_override(
            "sandbox_workspace_write.network_access",
            network_access_enabled,
        )

    def _add_web_search_mode(self) -> None:
        self._seen_args.add("web_search_mode")
        self._seen_args.add("web_search_enabled")

        web_search_mode = self._args.get("web_search_mode")
        if web_search_mode:
            self._append_config_override("web_search", web_search_mode)
            return

        web_search_enabled = self._args.get("web_search_enabled")
        if web_search_enabled is True:
            self._append_config_override("web_search", "live")
        elif web_search_enabled is False:
            self._append_config_override("web_search", "disabled")

    def _add_approval_policy(self) -> None:
        self._seen_args.add("approval_policy")

        approval_policy = self._args.get("approval_policy")
        if not approval_policy:
            return

        self._append_config_override("approval_policy", approval_policy)

    def _add_thread_id(self) -> None:
        self._seen_args.add("thread_id")

        thread_id = self._args.get("thread_id")
        if not thread_id:
            return

        self._command.argv.append("resume")
        self._command.argv.append(thread_id)

    def _add_images(self) -> None:
        self._seen_args.add("images")

        images = self._args.get("images")
        if not images:
            return

        for image in images:
            self._command.argv.append("--image")
            self._command.argv.append(image)

    def _append_config_override(self, key: str, value: CodexConfigValue) -> None:
        config_value = to_config_value(value, key)
        self._command.argv.append("--config")
        self._command.argv.append(f"{key}={config_value}")


def build_exec_args(  # noqa: PLR0913
    *,
    input: Input,  # noqa: A002
    options: CodexOptions,
    thread_options: ThreadOptions,
    thread_id: str | None,
    turn_options: TurnOptions,
    output_schema_path: str | None,
) -> CodexExecArgs:
    normalized = normalize_input(input)
    args = CodexExecArgs(
        input=normalized.prompt,
        images=normalized.images,
        **thread_options,  # type: ignore[typeddict-item]
    )

    base_url = options.get("base_url")
    if base_url is not None:
        args["base_url"] = base_url

    api_key = options.get("api_key")
    if api_key is not None:
        args["api_key"] = api_key

    if thread_id is not None:
        args["thread_id"] = thread_id

    signal = turn_options.get("signal")
    if signal is not None:
        args["signal"] = signal

    if output_schema_path is not None:
        args["output_schema_file"] = output_schema_path

    return args



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal, TypeAlias

from acodex.types.items import ThreadItem


@dataclass(frozen=True, slots=True)
class ThreadStartedEvent:
    """Emitted when a new thread is started as the first event."""

    thread_id: str
    """Identifier of the new thread. This can be used to resume the thread later."""

    type: Literal["thread.started"] = field(default="thread.started", init=False)
    """Discriminator with value ``"thread.started"``."""


@dataclass(frozen=True, slots=True)
class TurnStartedEvent:
    """Emitted when a turn is started by sending a new prompt to the model.

    A turn encompasses all events that happen while the agent is processing the prompt.
    """

    type: Literal["turn.started"] = field(default="turn.started", init=False)
    """Discriminator with value ``"turn.started"``."""


@dataclass(frozen=True, slots=True)
class Usage:
    """Describes token usage for a completed turn."""

    input_tokens: int
    """Number of input tokens used during the turn."""

    cached_input_tokens: int
    """Number of cached input tokens used during the turn."""

    output_tokens: int
    """Number of output tokens used during the turn."""


@dataclass(frozen=True, slots=True)
class TurnCompletedEvent:
    """Emitted when a turn is completed.

    This is typically emitted right after the assistant response.

    """

    usage: Usage
    """Token usage for the completed turn."""

    type: Literal["turn.completed"] = field(default="turn.completed", init=False)
    """Discriminator with value ``"turn.completed"``."""


@dataclass(frozen=True, slots=True)
class ThreadError:
    """Fatal error payload for a failed turn."""

    message: str
    """Error message describing the failure."""


@dataclass(frozen=True, slots=True)
class TurnFailedEvent:
    """Indicates that a turn failed with an error."""

    error: ThreadError
    """Fatal error payload for the failed turn."""

    type: Literal["turn.failed"] = field(default="turn.failed", init=False)
    """Discriminator with value ``"turn.failed"``."""


@dataclass(frozen=True, slots=True)
class ItemStartedEvent:
    """Emitted when a new item is added to the thread.

    Typically, the item is initially in progress.

    """

    item: ThreadItem
    """Thread item payload that was started."""

    type: Literal["item.started"] = field(default="item.started", init=False)
    """Discriminator with value ``"item.started"``."""


@dataclass(frozen=True, slots=True)
class ItemUpdatedEvent:
    """Emitted when an item is updated."""

    item: ThreadItem
    """Updated thread item payload."""

    type: Literal["item.updated"] = field(default="item.updated", init=False)
    """Discriminator with value ``"item.updated"``."""


@dataclass(frozen=True, slots=True)
class ItemCompletedEvent:
    """Signals that an item has reached a terminal state.

    The terminal state may be success or failure depending on the item payload.

    """

    item: ThreadItem
    """Thread item payload in a terminal state."""

    type: Literal["item.completed"] = field(default="item.completed", init=False)
    """Discriminator with value ``"item.completed"``."""


@dataclass(frozen=True, slots=True)
class ThreadErrorEvent:
    """Represents an unrecoverable error emitted directly by the event stream."""

    message: str
    """Error message emitted by the stream."""

    type: Literal["error"] = field(default="error", init=False)
    """Discriminator with value ``"error"``."""


ThreadEvent: TypeAlias = (
    ThreadStartedEvent
    | TurnStartedEvent
    | TurnCompletedEvent
    | TurnFailedEvent
    | ItemStartedEvent
    | ItemUpdatedEvent
    | ItemCompletedEvent
    | ThreadErrorEvent
)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypeAlias, TypedDict

if TYPE_CHECKING:
    from typing_extensions import NotRequired

ApprovalMode: TypeAlias = Literal["never", "on-request", "on-failure", "untrusted"]
SandboxMode: TypeAlias = Literal["read-only", "workspace-write", "danger-full-access"]
ModelReasoningEffort: TypeAlias = Literal["minimal", "low", "medium", "high", "xhigh"]
WebSearchMode: TypeAlias = Literal["disabled", "cached", "live"]


class ThreadOptions(TypedDict):
    """Thread-level options applied to each turn."""

    model: NotRequired[str]
    """Model name override."""

    sandbox_mode: NotRequired[SandboxMode]
    """Sandbox mode for command execution."""

    working_directory: NotRequired[str]
    """Working directory used for the turn."""

    skip_git_repo_check: NotRequired[bool]
    """Whether to skip repository checks."""

    model_reasoning_effort: NotRequired[ModelReasoningEffort]
    """Requested model reasoning effort."""

    network_access_enabled: NotRequired[bool]
    """Whether network access is enabled in sandboxed execution."""

    web_search_mode: NotRequired[WebSearchMode]
    """Web search mode configuration."""

    web_search_enabled: NotRequired[bool]
    """Legacy web search enable/disable flag."""

    approval_policy: NotRequired[ApprovalMode]
    """Approval policy for tool and command execution."""

    additional_directories: NotRequired[list[str]]
    """Additional directories to add to the execution workspace."""



from __future__ import annotations

from dataclasses import MISSING, Field, fields, is_dataclass
from typing import Literal, get_args, get_origin, get_type_hints

from acodex.types import events as py_events, items as py_items
from tests.compatibility.vendor_ts_sdk import VENDOR_TS_SDK_SRC
from tools.compatibility.ts_type_alias_parser import TsAlias, parse_exported_type_aliases


def test_events_aliases_exist_in_python_sdk() -> None:
    ts_aliases = _load_event_aliases()
    missing_aliases = sorted(
        alias_name for alias_name in ts_aliases if not hasattr(py_events, alias_name)
    )

    assert not missing_aliases, "Missing event aliases from Python SDK: " + ", ".join(
        missing_aliases,
    )


def test_events_object_aliases_match_dataclass_shapes() -> None:
    ts_aliases = _load_event_aliases()

    for alias_name, ts_alias in ts_aliases.items():
        if ts_alias.kind != "object":
            continue

        py_symbol = getattr(py_events, alias_name)
        assert is_dataclass(py_symbol), f"{alias_name} must be a dataclass in Python SDK"

        ts_properties = ts_alias.properties
        assert ts_properties is not None, f"{alias_name} object alias must expose parsed properties"

        py_fields = {field_info.name: field_info for field_info in fields(py_symbol)}
        assert set(ts_properties) == set(py_fields), (
            f"{alias_name} fields mismatch: TS={sorted(ts_properties)}, Python={sorted(py_fields)}"
        )

        type_hints = get_type_hints(py_symbol, include_extras=True)
        for property_name, ts_property in ts_properties.items():
            py_field = py_fields[property_name]

            if property_name == "type":
                expected_discriminator = _parse_string_literal(ts_property.type_expr)
                assert expected_discriminator is not None, (
                    f"{alias_name}.type must be a string literal in TypeScript"
                )
                assert py_field.default == expected_discriminator, (
                    f"{alias_name}.type default mismatch: "
                    f"TS={expected_discriminator!r}, Python={py_field.default!r}"
                )
                python_type_hint = type_hints["type"]
                assert get_origin(python_type_hint) is Literal, (
                    f"{alias_name}.type must be annotated as typing.Literal in Python SDK"
                )
                assert get_args(python_type_hint) == (expected_discriminator,), (
                    f"{alias_name}.type literal annotation mismatch: "
                    f"TS={expected_discriminator!r}, Python={get_args(python_type_hint)!r}"
                )
                continue

            if ts_property.optional:
                assert not _is_required_dataclass_field(py_field), (
                    f"{alias_name}.{property_name} is optional in TS but required in Python"
                )
            else:
                assert _is_required_dataclass_field(py_field), (
                    f"{alias_name}.{property_name} is required in TS but optional in Python"
                )


def test_events_cross_type_references_match_python_types() -> None:
    turn_completed_hints = get_type_hints(py_events.TurnCompletedEvent, include_extras=True)
    turn_failed_hints = get_type_hints(py_events.TurnFailedEvent, include_extras=True)

    assert turn_completed_hints["usage"] is py_events.Usage, (
        "TurnCompletedEvent.usage must resolve to acodex.types.events.Usage"
    )
    assert turn_failed_hints["error"] is py_events.ThreadError, (
        "TurnFailedEvent.error must resolve to acodex.types.events.ThreadError"
    )

    for item_event_type in (
        py_events.ItemStartedEvent,
        py_events.ItemUpdatedEvent,
        py_events.ItemCompletedEvent,
    ):
        item_hints = get_type_hints(item_event_type, include_extras=True)
        assert item_hints["item"] == py_items.ThreadItem, (
            f"{item_event_type.__name__}.item must resolve to acodex.types.items.ThreadItem"
        )


def test_thread_event_union_membership_matches_typescript() -> None:
    ts_alias = _load_event_aliases()["ThreadEvent"]
    assert ts_alias.kind == "identifier_union", (
        "ThreadEvent in TypeScript must be an identifier union"
    )
    ts_members = set(ts_alias.union_members or ())
    py_members = {member.__name__ for member in get_args(py_events.ThreadEvent)}

    assert ts_members == py_members, (
        f"ThreadEvent union mismatch: TS={sorted(ts_members)}, Python={sorted(py_members)}"
    )


def _load_event_aliases() -> dict[str, TsAlias]:
    events_ts_path = VENDOR_TS_SDK_SRC / "events.ts"
    return parse_exported_type_aliases(events_ts_path.read_text(encoding="utf-8"))


def _is_required_dataclass_field(field_info: Field[object]) -> bool:
    return field_info.default is MISSING and field_info.default_factory is MISSING


def _parse_string_literal(type_expr: str) -> str | None:
    stripped_type_expr = type_expr.strip()
    if len(stripped_type_expr) < 2:
        return None

    quote = stripped_type_expr[0]
    if quote not in {'"', "'"} or stripped_type_expr[-1] != quote:
        return None

    return stripped_type_expr[1:-1]



from __future__ import annotations

from collections.abc import AsyncIterator, Iterator

from acodex._internal.exec import CodexExecArgs, CodexExecCLICommandBuilder
from acodex._internal.locator import find_codex_path
from acodex._internal.process_runner import AsyncCodexProcessRunner, SyncCodexProcessRunner
from acodex.types.codex_options import CodexConfigObject


class CodexExec:
    """Execute `codex exec` calls for the synchronous client."""

    def __init__(
        self,
        *,
        executable_path: str | None = None,
        env: dict[str, str] | None = None,
        config_overrides: CodexConfigObject | None = None,
    ) -> None:
        self._executable_path = executable_path or find_codex_path()
        self._env = env
        self._config_overrides = config_overrides

    def run(self, args: CodexExecArgs) -> Iterator[str]:
        """Run Codex and stream JSONL lines.

        Yields:
            JSONL line payloads emitted by the Codex CLI.

        """
        command = CodexExecCLICommandBuilder(
            args=args,
            config_overrides=self._config_overrides,
            env_overrides=self._env,
        ).build_command()
        runner = SyncCodexProcessRunner(executable_path=self._executable_path)
        yield from runner.stream_lines(command)


class AsyncCodexExec:
    """Execute `codex exec` calls for the asynchronous client."""

    def __init__(
        self,
        *,
        executable_path: str | None = None,
        env: dict[str, str] | None = None,
        config_overrides: CodexConfigObject | None = None,
    ) -> None:
        self._executable_path = executable_path or find_codex_path()
        self._env = env
        self._config_overrides = config_overrides

    async def run(self, args: CodexExecArgs) -> AsyncIterator[str]:
        """Run Codex and stream JSONL lines asynchronously.

        Yields:
            JSONL line payloads emitted by the Codex CLI.

        """
        command = CodexExecCLICommandBuilder(
            args=args,
            config_overrides=self._config_overrides,
            env_overrides=self._env,
        ).build_command()
        runner = AsyncCodexProcessRunner(executable_path=self._executable_path)
        async for line in runner.stream_lines(command):
            yield line



from __future__ import annotations

from acodex.codex import AsyncCodex, Codex
from acodex.thread import AsyncThread, Thread
from acodex.types.codex_options import CodexOptions
from acodex.types.events import (
    ItemCompletedEvent,
    ItemStartedEvent,
    ItemUpdatedEvent,
    ThreadError,
    ThreadErrorEvent,
    ThreadEvent,
    ThreadStartedEvent,
    TurnCompletedEvent,
    TurnFailedEvent,
    TurnStartedEvent,
    Usage,
)
from acodex.types.input import Input, UserInput
from acodex.types.items import (
    AgentMessageItem,
    CommandExecutionItem,
    ErrorItem,
    FileChangeItem,
    McpToolCallItem,
    ReasoningItem,
    ThreadItem,
    TodoListItem,
    WebSearchItem,
)
from acodex.types.thread_options import (
    ApprovalMode,
    ModelReasoningEffort,
    SandboxMode,
    ThreadOptions,
    WebSearchMode,
)
from acodex.types.turn import RunResult, RunStreamedResult
from acodex.types.turn_options import TurnOptions

__all__ = [
    "AgentMessageItem",
    "ApprovalMode",
    "AsyncCodex",
    "AsyncThread",
    "Codex",
    "CodexOptions",
    "CommandExecutionItem",
    "ErrorItem",
    "FileChangeItem",
    "Input",
    "ItemCompletedEvent",
    "ItemStartedEvent",
    "ItemUpdatedEvent",
    "McpToolCallItem",
    "ModelReasoningEffort",
    "ReasoningItem",
    "RunResult",
    "RunStreamedResult",
    "SandboxMode",
    "Thread",
    "ThreadError",
    "ThreadErrorEvent",
    "ThreadEvent",
    "ThreadItem",
    "ThreadOptions",
    "ThreadStartedEvent",
    "TodoListItem",
    "TurnCompletedEvent",
    "TurnFailedEvent",
    "TurnOptions",
    "TurnStartedEvent",
    "Usage",
    "UserInput",
    "WebSearchItem",
    "WebSearchMode",
]



from __future__ import annotations

from dataclasses import MISSING, Field, fields, is_dataclass
from typing import Literal, get_args, get_origin, get_type_hints

from acodex.types import items as py_items
from tests.compatibility.vendor_ts_sdk import VENDOR_TS_SDK_SRC
from tools.compatibility.ts_type_alias_parser import (
    TsAlias,
    TsProperty,
    parse_exported_type_aliases,
)


def test_items_aliases_exist_in_python_sdk() -> None:
    ts_aliases = _load_item_aliases()
    missing_aliases = sorted(
        alias_name for alias_name in ts_aliases if not hasattr(py_items, alias_name)
    )

    assert not missing_aliases, "Missing item aliases from Python SDK: " + ", ".join(
        missing_aliases,
    )


def test_item_string_literal_unions_match_typescript() -> None:
    ts_aliases = _load_item_aliases()
    literal_union_aliases = (
        "CommandExecutionStatus",
        "PatchChangeKind",
        "PatchApplyStatus",
        "McpToolCallStatus",
    )

    for alias_name in literal_union_aliases:
        ts_alias = ts_aliases[alias_name]
        assert ts_alias.kind == "string_union", (
            f"{alias_name} in TypeScript must be a string-literal union"
        )
        ts_literals = set(ts_alias.string_literals or ())
        py_literals = set(get_args(getattr(py_items, alias_name)))
        assert ts_literals == py_literals, (
            f"{alias_name} literal union mismatch: TS={sorted(ts_literals)}, "
            f"Python={sorted(py_literals)}"
        )


def test_item_object_aliases_match_dataclass_shapes() -> None:
    ts_aliases = _load_item_aliases()

    for alias_name, ts_alias in ts_aliases.items():
        if ts_alias.kind != "object":
            continue

        py_symbol = getattr(py_items, alias_name)
        assert is_dataclass(py_symbol), f"{alias_name} must be a dataclass in Python SDK"

        ts_properties = ts_alias.properties
        assert ts_properties is not None, f"{alias_name} object alias must expose parsed properties"

        py_fields = {field_info.name: field_info for field_info in fields(py_symbol)}
        assert set(ts_properties) == set(py_fields), (
            f"{alias_name} fields mismatch: TS={sorted(ts_properties)}, Python={sorted(py_fields)}"
        )

        type_hints = get_type_hints(py_symbol, include_extras=True)
        for property_name, ts_property in ts_properties.items():
            py_field = py_fields[property_name]

            if property_name == "type":
                expected_discriminator = _parse_string_literal(ts_property.type_expr)
                assert expected_discriminator is not None, (
                    f"{alias_name}.type must be a string literal in TypeScript"
                )
                assert py_field.default == expected_discriminator, (
                    f"{alias_name}.type default mismatch: "
                    f"TS={expected_discriminator!r}, Python={py_field.default!r}"
                )
                python_type_hint = type_hints["type"]
                assert get_origin(python_type_hint) is Literal, (
                    f"{alias_name}.type must be annotated as typing.Literal in Python SDK"
                )
                assert get_args(python_type_hint) == (expected_discriminator,), (
                    f"{alias_name}.type literal annotation mismatch: "
                    f"TS={expected_discriminator!r}, Python={get_args(python_type_hint)!r}"
                )
                continue

            if ts_property.optional:
                assert not _is_required_dataclass_field(py_field), (
                    f"{alias_name}.{property_name} is optional in TS but required in Python"
                )
            else:
                assert _is_required_dataclass_field(py_field), (
                    f"{alias_name}.{property_name} is required in TS but optional in Python"
                )


def test_optional_item_fields_match_typescript() -> None:
    ts_aliases = _load_item_aliases()
    optional_expectations: tuple[tuple[str, str], ...] = (
        ("CommandExecutionItem", "exit_code"),
        ("McpToolCallItem", "result"),
        ("McpToolCallItem", "error"),
    )

    for alias_name, property_name in optional_expectations:
        ts_property = _require_property(
            ts_aliases,
            alias_name=alias_name,
            property_name=property_name,
        )
        assert ts_property.optional, f"{alias_name}.{property_name} must be optional in TypeScript"

        py_field_map = {
            field_info.name: field_info for field_info in fields(getattr(py_items, alias_name))
        }
        assert not _is_required_dataclass_field(py_field_map[property_name]), (
            f"{alias_name}.{property_name} must be optional in Python"
        )


def test_mcp_tool_call_inline_object_shapes_match_python_helpers() -> None:
    ts_aliases = _load_item_aliases()
    ts_mcp_tool_call_alias = ts_aliases["McpToolCallItem"]
    ts_properties = ts_mcp_tool_call_alias.properties
    assert ts_properties is not None, "McpToolCallItem must expose parsed properties"

    ts_result_property = _require_property(
        ts_aliases,
        alias_name="McpToolCallItem",
        property_name="result",
    )
    ts_error_property = _require_property(
        ts_aliases,
        alias_name="McpToolCallItem",
        property_name="error",
    )

    assert ts_result_property.inline_object is not None, (
        "McpToolCallItem.result must be parsed as an inline TypeScript object"
    )
    assert ts_error_property.inline_object is not None, (
        "McpToolCallItem.error must be parsed as an inline TypeScript object"
    )

    python_hints = get_type_hints(py_items.McpToolCallItem, include_extras=True)
    python_result_type = _get_optional_union_member(python_hints["result"])
    python_error_type = _get_optional_union_member(python_hints["error"])

    assert python_result_type is py_items.McpToolCallResult, (
        "McpToolCallItem.result must resolve to acodex.types.items.McpToolCallResult | None"
    )
    assert python_error_type is py_items.McpToolCallError, (
        "McpToolCallItem.error must resolve to acodex.types.items.McpToolCallError | None"
    )

    ts_result_fields = set((ts_result_property.inline_object.properties or {}).keys())
    ts_error_fields = set((ts_error_property.inline_object.properties or {}).keys())
    py_result_fields = {field_info.name for field_info in fields(py_items.McpToolCallResult)}
    py_error_fields = {field_info.name for field_info in fields(py_items.McpToolCallError)}

    assert ts_result_fields == py_result_fields, (
        "McpToolCallResult field mismatch: "
        f"TS={sorted(ts_result_fields)}, Python={sorted(py_result_fields)}"
    )
    assert ts_error_fields == py_error_fields, (
        "McpToolCallError field mismatch: "
        f"TS={sorted(ts_error_fields)}, Python={sorted(py_error_fields)}"
    )


def test_thread_item_union_membership_matches_typescript() -> None:
    ts_alias = _load_item_aliases()["ThreadItem"]
    assert ts_alias.kind == "identifier_union", (
        "ThreadItem in TypeScript must be an identifier union"
    )
    ts_members = set(ts_alias.union_members or ())
    py_members = {member.__name__ for member in get_args(py_items.ThreadItem)}

    assert ts_members == py_members, (
        f"ThreadItem union mismatch: TS={sorted(ts_members)}, Python={sorted(py_members)}"
    )


def _load_item_aliases() -> dict[str, TsAlias]:
    items_ts_path = VENDOR_TS_SDK_SRC / "items.ts"
    return parse_exported_type_aliases(items_ts_path.read_text(encoding="utf-8"))


def _is_required_dataclass_field(field_info: Field[object]) -> bool:
    return field_info.default is MISSING and field_info.default_factory is MISSING


def _parse_string_literal(type_expr: str) -> str | None:
    stripped_type_expr = type_expr.strip()
    if len(stripped_type_expr) < 2:
        return None

    quote = stripped_type_expr[0]
    if quote not in {'"', "'"} or stripped_type_expr[-1] != quote:
        return None

    return stripped_type_expr[1:-1]


def _require_property(
    aliases: dict[str, TsAlias],
    *,
    alias_name: str,
    property_name: str,
) -> TsProperty:
    ts_alias = aliases[alias_name]
    ts_properties = ts_alias.properties
    assert ts_properties is not None, f"{alias_name} must expose parsed object properties"
    assert property_name in ts_properties, (
        f"{alias_name} missing expected property: {property_name}"
    )
    return ts_properties[property_name]


def _get_optional_union_member(annotation: object) -> type[object]:
    non_none_members = [member for member in get_args(annotation) if member is not type(None)]
    assert len(non_none_members) == 1, (
        f"Expected an optional union with one concrete member, got: {annotation!r}"
    )

    union_member = non_none_members[0]
    assert isinstance(union_member, type), f"Expected type member, got: {union_member!r}"
    return union_member



from __future__ import annotations


class CodexError(Exception):
    """Base exception for all errors raised by acodex."""


class CodexCancelledError(CodexError):
    """Raised when a turn is canceled via TurnOptions.signal."""


class CodexExecError(CodexError):
    """Raised when the codex executable fails to start or exits unsuccessfully."""

    def __init__(self, message: str, *, stderr: str | None = None) -> None:
        super().__init__(message)
        self.stderr = stderr


class CodexExecutableNotFoundError(CodexExecError):
    """Raised when the codex executable cannot be discovered on PATH."""

    def __init__(
        self,
        message: str,
        *,
        executable_name: str = "codex",
    ) -> None:
        super().__init__(message)
        self.executable_name = executable_name


class CodexThreadRunError(CodexError):
    """Raised when a thread run fails while streaming or parsing events."""


class CodexConfigError(CodexError):
    """Raised when Codex config overrides are invalid or cannot be serialized."""


class CodexOutputSchemaError(CodexError):
    """Raised when a TurnOptions.output_schema payload is invalid."""


class CodexInternalError(CodexError):
    """Raised when an internal invariant is violated, indicating a bug in acodex."""



from __future__ import annotations

import os
import re
from pathlib import Path

import pytest

from acodex._internal.locator import find_codex_path
from acodex.exceptions import CodexExecutableNotFoundError
from acodex.exec import AsyncCodexExec, CodexExec


def test_find_codex_path_resolves_executable_from_path(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    executable = _create_fake_codex_executable(monkeypatch, tmp_path)
    monkeypatch.setenv("PATH", str(tmp_path))

    found = find_codex_path()

    assert Path(found).resolve() == executable.resolve()


def test_find_codex_path_raises_when_missing(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("PATH", "")

    with pytest.raises(
        CodexExecutableNotFoundError,
        match=re.escape(
            'Could not locate "codex" on PATH. Install the Codex CLI or pass `codex_path_override`.',
        ),
    ) as error:
        find_codex_path()

    assert error.value.executable_name == "codex"


def test_exec_constructors_resolve_executable_path_from_locator(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    executable = _create_fake_codex_executable(monkeypatch, tmp_path)
    monkeypatch.setenv("PATH", str(tmp_path))

    sync_exec = CodexExec()
    async_exec = AsyncCodexExec()
    expected_path = str(executable)

    assert sync_exec._executable_path == expected_path
    assert async_exec._executable_path == expected_path


def test_exec_constructors_keep_explicit_executable_path_when_path_is_empty(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("PATH", "")
    explicit_path = "explicit/path/to/codex"

    sync_exec = CodexExec(executable_path=explicit_path)
    async_exec = AsyncCodexExec(executable_path=explicit_path)

    assert sync_exec._executable_path == explicit_path
    assert async_exec._executable_path == explicit_path


def test_exec_constructors_treat_empty_string_as_discovery(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    executable = _create_fake_codex_executable(monkeypatch, tmp_path)
    monkeypatch.setenv("PATH", str(tmp_path))

    sync_exec = CodexExec(executable_path="")
    async_exec = AsyncCodexExec(executable_path="")
    expected_path = str(executable)

    assert sync_exec._executable_path == expected_path
    assert async_exec._executable_path == expected_path


def _create_fake_codex_executable(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> Path:
    if os.name == "nt":
        monkeypatch.setenv("PATHEXT", ".EXE;.BAT;.CMD")
        executable = tmp_path / "codex.bat"
        executable.write_text("@echo off\r\nexit /B 0\r\n", encoding="utf-8")
        return executable

    executable = tmp_path / "codex"
    executable.write_text("#!/bin/sh\nexit 0\n", encoding="utf-8")
    executable.chmod(0o755)
    return executable



from __future__ import annotations

from collections.abc import AsyncIterator, Iterator
from dataclasses import dataclass
from typing import Generic, TypeAlias, TypeVar

from acodex.types.events import ThreadEvent, Usage
from acodex.types.items import ThreadItem

EventsT = TypeVar("EventsT")


@dataclass(frozen=True, slots=True)
class Turn:
    """Completed turn.

    This is the result returned by `run()`.
    """

    items: list[ThreadItem]
    final_response: str
    usage: Usage | None


@dataclass(frozen=True, slots=True)
class StreamedTurn(Generic[EventsT]):
    """The result of the `run_streamed` method."""

    events: EventsT


RunResult: TypeAlias = Turn

RunStreamedResult: TypeAlias = StreamedTurn[Iterator[ThreadEvent]]
AsyncRunStreamedResult: TypeAlias = StreamedTurn[AsyncIterator[ThreadEvent]]



from __future__ import annotations

import asyncio
import json
from pathlib import Path

import pytest

from acodex.exec import AsyncCodexExec, CodexExec
from tests.unit.fake_codex_executable import create_fake_codex_executable


def test_codex_exec_run_streams_stdout_lines(tmp_path: Path) -> None:
    executable = create_fake_codex_executable(tmp_path)
    client = CodexExec(
        executable_path=str(executable),
        env={
            "FAKE_CODEX_MODE": "lines",
            "FAKE_LINES_JSON": json.dumps(["first", "second", "third"]),
        },
    )

    lines = list(client.run({"input": "hello"}))

    assert lines == ["first", "second", "third"]


def test_async_codex_exec_run_streams_stdout_lines(tmp_path: Path) -> None:
    executable = create_fake_codex_executable(tmp_path)
    client = AsyncCodexExec(
        executable_path=str(executable),
        env={
            "FAKE_CODEX_MODE": "lines",
            "FAKE_LINES_JSON": json.dumps(["async-first", "async-second"]),
        },
    )

    async def run() -> list[str]:
        return [line async for line in client.run({"input": "hello"})]

    assert asyncio.run(run()) == ["async-first", "async-second"]


def test_exec_env_empty_overrides_prevent_parent_env_inheritance(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    executable = create_fake_codex_executable(tmp_path)
    marker_name = "ACODEX_PARENT_MARKER"
    monkeypatch.setenv(marker_name, "1")
    client = CodexExec(
        executable_path=str(executable),
        env={
            "FAKE_CODEX_MODE": "env_guard",
            "FAKE_MARKER_NAME": marker_name,
        },
    )

    lines = list(client.run({"input": "hello"}))

    assert lines == ["clean"]



from __future__ import annotations

import hashlib
import json
import os
import re
from datetime import datetime, timezone
from operator import itemgetter
from pathlib import Path

from tests.compatibility.vendor_ts_sdk import VENDOR_TS_SDK_ROOT


def _compute_tree_hash(*, root_dir: Path) -> str:
    upstream_json = root_dir / "UPSTREAM.json"
    hasher = hashlib.sha256()

    entries: list[tuple[str, Path]] = []
    for dirpath, _, filenames in os.walk(root_dir):
        for filename in filenames:
            path = Path(dirpath) / filename
            if not path.is_file():
                continue
            if path == upstream_json:
                continue
            rel_posix = path.relative_to(root_dir).as_posix()
            entries.append((rel_posix, path))

    entries.sort(key=itemgetter(0))

    for rel_posix, path in entries:
        hasher.update(rel_posix.encode("utf-8"))
        hasher.update(b"\0")
        hasher.update(path.read_bytes())

    return f"sha256:{hasher.hexdigest()}"


def _parse_github_iso8601(value: str) -> datetime:
    # GitHub typically returns timestamps like "2026-02-12T20:05:52Z".
    normalized = value.replace("Z", "+00:00")
    parsed = datetime.fromisoformat(normalized)
    if parsed.tzinfo is None:
        parsed = parsed.replace(tzinfo=timezone.utc)
    return parsed


def test_vendor_metadata_file_exists_and_is_valid() -> None:
    upstream_path = VENDOR_TS_SDK_ROOT / "UPSTREAM.json"
    assert upstream_path.exists(), "Missing vendor/codex-ts-sdk/UPSTREAM.json"

    data = json.loads(upstream_path.read_text(encoding="utf-8"))
    assert isinstance(data, dict), "UPSTREAM.json must be a JSON object"

    assert data.get("repo") == "https://github.com/openai/codex"
    assert data.get("path") == "sdk/typescript"

    release_tag = data.get("release_tag")
    assert isinstance(release_tag, str)
    assert re.match(r"^rust-v\d+\.\d+\.\d+$", release_tag)

    release_id = data.get("release_id")
    assert isinstance(release_id, int)
    assert release_id > 0

    release_published_at = data.get("release_published_at")
    assert isinstance(release_published_at, str)
    assert release_published_at
    _parse_github_iso8601(release_published_at)

    release_url = data.get("release_url")
    assert isinstance(release_url, str)
    assert release_url

    commit = data.get("commit")
    assert isinstance(commit, str)
    assert re.match(r"^[0-9a-f]{40}$", commit)

    tree_hash = data.get("tree_hash")
    assert isinstance(tree_hash, str)
    assert re.match(r"^sha256:[0-9a-f]{64}$", tree_hash)
    assert tree_hash == _compute_tree_hash(root_dir=VENDOR_TS_SDK_ROOT)


def test_vendor_required_files_exist() -> None:
    required_paths = (
        VENDOR_TS_SDK_ROOT / "package.json",
        VENDOR_TS_SDK_ROOT / "src" / "items.ts",
        VENDOR_TS_SDK_ROOT / "src" / "events.ts",
        VENDOR_TS_SDK_ROOT / "src" / "index.ts",
    )
    for path in required_paths:
        assert path.exists(), f"Missing required vendored file: {path}"


def test_package_json_points_at_upstream_repo_and_directory() -> None:
    package_json_path = VENDOR_TS_SDK_ROOT / "package.json"
    data = json.loads(package_json_path.read_text(encoding="utf-8"))
    assert isinstance(data, dict), "vendor/codex-ts-sdk/package.json must be a JSON object"

    repository = data.get("repository")
    assert isinstance(repository, dict), "package.json repository must be an object"

    url = repository.get("url")
    assert isinstance(url, str)
    assert "openai/codex" in url

    directory = repository.get("directory")
    assert directory == "sdk/typescript"



from __future__ import annotations

import argparse
import hashlib
import json
import os
import shutil
import subprocess  # noqa: S404
import sys
import tempfile
import urllib.error
import urllib.parse
import urllib.request
from operator import itemgetter
from pathlib import Path
from typing import Final

_REPO_URL: Final[str] = "https://github.com/openai/codex"
_REPO_GIT_URL: Final[str] = "https://github.com/openai/codex.git"
_REPO_SLUG: Final[str] = "openai/codex"
_UPSTREAM_PATH: Final[str] = "sdk/typescript"
_API_ROOT: Final[str] = "https://api.github.com"
_USER_AGENT: Final[str] = "acodex-vendor-bot"
_HTTP_NOT_FOUND: Final[int] = 404
_GIT_SHA_LEN: Final[int] = 40


class _VendorError(RuntimeError):
    pass


def _repo_root() -> Path:
    return Path(__file__).resolve().parents[2]


def _vendor_root() -> Path:
    return _repo_root() / "vendor" / "codex-ts-sdk"


def _upstream_json_path() -> Path:
    return _vendor_root() / "UPSTREAM.json"


def _github_get_json(*, url: str) -> object:
    headers: dict[str, str] = {
        "Accept": "application/vnd.github+json",
        "User-Agent": _USER_AGENT,
    }

    token = os.environ.get("GITHUB_TOKEN")
    if token:
        headers["Authorization"] = f"Bearer {token}"

    request = urllib.request.Request(url, headers=headers)  # noqa: S310
    try:
        with urllib.request.urlopen(request, timeout=30) as response:  # noqa: S310
            body = response.read().decode("utf-8")
    except urllib.error.HTTPError as exc:
        if exc.code == _HTTP_NOT_FOUND:
            raise _VendorError(f"Tag is not an official GitHub Release: {url}") from exc
        raise _VendorError(f"GitHub API error: {exc.code} {exc.reason} for {url}") from exc
    except urllib.error.URLError as exc:  # pragma: no cover
        raise _VendorError(f"GitHub API request failed for {url}: {exc.reason}") from exc

    data: object = json.loads(body)
    return data


def _require_dict(value: object, *, context: str) -> dict[str, object]:
    if not isinstance(value, dict):
        raise _VendorError(f"Expected object for {context}, got {type(value).__name__}")
    return value


def _require_str(obj: dict[str, object], key: str, *, context: str) -> str:
    value = obj.get(key)
    if not isinstance(value, str) or not value:
        raise _VendorError(f"Missing/invalid {context}.{key}")
    return value


def _require_int(obj: dict[str, object], key: str, *, context: str) -> int:
    value = obj.get(key)
    if not isinstance(value, int) or value <= 0:
        raise _VendorError(f"Missing/invalid {context}.{key}")
    return value


def _require_bool(obj: dict[str, object], key: str, *, context: str) -> bool:
    value = obj.get(key)
    if not isinstance(value, bool):
        raise _VendorError(f"Missing/invalid {context}.{key}")
    return value


def _git_executable() -> str:
    git = shutil.which("git")
    if git is None:
        raise _VendorError("git not found on PATH.")
    return git


def _run_git(*args: str, cwd: Path) -> subprocess.CompletedProcess[str]:
    return subprocess.run(  # noqa: S603
        [_git_executable(), *args],
        cwd=str(cwd),
        check=False,
        text=True,
        capture_output=True,
    )


def _git_check_object(*, repo_dir: Path, revspec: str) -> bool:
    proc = _run_git("cat-file", "-e", revspec, cwd=repo_dir)
    return proc.returncode == 0


def _git_show_bytes(*, repo_dir: Path, revspec: str) -> bytes:
    proc = subprocess.run(  # noqa: S603
        [_git_executable(), "-C", str(repo_dir), "show", revspec],
        check=False,
        capture_output=True,
    )
    if proc.returncode != 0:
        stderr = proc.stderr.decode("utf-8", errors="replace")
        raise _VendorError(f"git show failed for {revspec}: {stderr.strip()}")
    return proc.stdout


def _git_rev_parse_head(*, repo_dir: Path) -> str:
    proc = _run_git("rev-parse", "HEAD", cwd=repo_dir)
    if proc.returncode != 0:
        raise _VendorError(f"git rev-parse HEAD failed: {proc.stderr.strip()}")
    sha = proc.stdout.strip()
    if len(sha) != _GIT_SHA_LEN or any(ch not in "0123456789abcdef" for ch in sha):
        raise _VendorError(f"Unexpected HEAD sha: {sha!r}")
    return sha


def _replace_tree(*, src_dir: Path, dst_dir: Path) -> None:
    if dst_dir.exists():
        shutil.rmtree(dst_dir)
    dst_dir.parent.mkdir(parents=True, exist_ok=True)
    shutil.copytree(src_dir, dst_dir)


def _tree_hash(*, root_dir: Path) -> str:
    upstream_json = root_dir / "UPSTREAM.json"
    hasher = hashlib.sha256()

    entries: list[tuple[str, Path]] = []
    for dirpath, _, filenames in os.walk(root_dir):
        for filename in filenames:
            path = Path(dirpath) / filename
            if not path.is_file():
                continue
            if path == upstream_json:
                continue
            rel_posix = path.relative_to(root_dir).as_posix()
            entries.append((rel_posix, path))

    entries.sort(key=itemgetter(0))

    for rel_posix, path in entries:
        hasher.update(rel_posix.encode("utf-8"))
        hasher.update(b"\0")
        hasher.update(path.read_bytes())

    return f"sha256:{hasher.hexdigest()}"


def _read_existing_tag() -> str:
    path = _upstream_json_path()
    if not path.exists():
        raise _VendorError("Missing vendor/codex-ts-sdk/UPSTREAM.json; pass --tag explicitly.")

    data: object = json.loads(path.read_text(encoding="utf-8"))
    obj = _require_dict(data, context="UPSTREAM.json")
    tag = obj.get("release_tag")
    if not isinstance(tag, str) or not tag:
        raise _VendorError("UPSTREAM.json missing/invalid release_tag.")
    return tag


def _resolve_release(*, tag: str) -> dict[str, object]:
    url_tag = urllib.parse.quote(tag, safe="")
    url = f"{_API_ROOT}/repos/{_REPO_SLUG}/releases/tags/{url_tag}"
    payload = _github_get_json(url=url)
    release = _require_dict(payload, context="release")

    if _require_bool(release, "draft", context="release"):
        raise _VendorError(f"Tag is a draft release (unsupported): {tag}")
    if _require_bool(release, "prerelease", context="release"):
        raise _VendorError(f"Tag is a prerelease (unsupported): {tag}")

    return {
        "release_id": _require_int(release, "id", context="release"),
        "release_tag": _require_str(release, "tag_name", context="release"),
        "release_published_at": _require_str(release, "published_at", context="release"),
        "release_url": _require_str(release, "html_url", context="release"),
    }


def _clone_sparse_checkout(*, tag: str, tmp_dir: Path) -> Path:
    repo_dir = tmp_dir / "codex"
    proc = subprocess.run(  # noqa: S603
        [
            _git_executable(),
            "clone",
            "--depth",
            "1",
            "--filter=blob:none",
            "--sparse",
            "--branch",
            tag,
            _REPO_GIT_URL,
            str(repo_dir),
        ],
        check=False,
        text=True,
        capture_output=True,
    )
    if proc.returncode != 0:
        raise _VendorError(f"git clone failed: {proc.stderr.strip()}")

    proc = _run_git("sparse-checkout", "set", _UPSTREAM_PATH, cwd=repo_dir)
    if proc.returncode != 0:
        raise _VendorError(f"git sparse-checkout failed: {proc.stderr.strip()}")

    return repo_dir


def _write_root_file_if_present(*, repo_dir: Path, tag: str, name: str, dst_dir: Path) -> None:
    revspec = f"{tag}:{name}"
    if not _git_check_object(repo_dir=repo_dir, revspec=revspec):
        return
    (dst_dir / name).write_bytes(_git_show_bytes(repo_dir=repo_dir, revspec=revspec))


def _parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=(
            "Vendor the Codex TypeScript SDK (sdk/typescript) from an official GitHub Release."
        ),
    )
    parser.add_argument(
        "--tag",
        help=(
            "Upstream release tag to vendor (must exist as a stable GitHub Release). "
            "If omitted, read release_tag from vendor/codex-ts-sdk/UPSTREAM.json."
        ),
    )
    return parser.parse_args(argv)


def main(argv: list[str]) -> int:
    args = _parse_args(argv)
    tag: str = args.tag or _read_existing_tag()

    release_info = _resolve_release(tag=tag)
    canonical_tag = release_info["release_tag"]
    if not isinstance(canonical_tag, str):
        raise _VendorError("Internal error: release_tag must be a string.")

    with tempfile.TemporaryDirectory(prefix="acodex-vendor-") as tmp:
        repo_dir = _clone_sparse_checkout(tag=canonical_tag, tmp_dir=Path(tmp))
        upstream_dir = repo_dir / _UPSTREAM_PATH
        if not upstream_dir.is_dir():
            raise _VendorError(f"Upstream path missing in {canonical_tag}: {_UPSTREAM_PATH}")

        vendor_dir = _vendor_root()
        _replace_tree(src_dir=upstream_dir, dst_dir=vendor_dir)

        _write_root_file_if_present(
            repo_dir=repo_dir,
            tag=canonical_tag,
            name="LICENSE",
            dst_dir=vendor_dir,
        )
        _write_root_file_if_present(
            repo_dir=repo_dir,
            tag=canonical_tag,
            name="NOTICE",
            dst_dir=vendor_dir,
        )

        commit = _git_rev_parse_head(repo_dir=repo_dir)
        tree_hash = _tree_hash(root_dir=vendor_dir)

    upstream_json = {
        "repo": _REPO_URL,
        "path": _UPSTREAM_PATH,
        "release_id": release_info["release_id"],
        "release_tag": canonical_tag,
        "release_published_at": release_info["release_published_at"],
        "release_url": release_info["release_url"],
        "commit": commit,
        "tree_hash": tree_hash,
    }

    _upstream_json_path().write_text(
        json.dumps(upstream_json, indent=2, sort_keys=True) + "\n",
        encoding="utf-8",
    )

    sys.stdout.write(f"Vendored {_UPSTREAM_PATH} at {canonical_tag} into vendor/codex-ts-sdk/\n")
    return 0


if __name__ == "__main__":
    try:
        raise SystemExit(main(sys.argv[1:]))
    except _VendorError as exc:
        sys.stderr.write(f"error: {exc}\n")
        raise SystemExit(1) from exc



from __future__ import annotations

import asyncio
import queue
import subprocess  # noqa: S404
import threading
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Iterator
from contextlib import suppress
from dataclasses import dataclass
from typing import IO, Final, NamedTuple, NoReturn, TypeAlias, cast

from acodex._internal.exec import CodexExecCommand
from acodex.exceptions import CodexCancelledError, CodexExecError

_READ_POLL_INTERVAL_SECONDS: Final[float] = 0.05
_PROCESS_TERMINATE_TIMEOUT_SECONDS: Final[float] = 1.0


class _StdoutEofSentinel:
    pass


class _AsyncTimeoutSentinel:
    pass


class _AsyncCancelledSentinel:
    pass


_STDOUT_EOF: Final[_StdoutEofSentinel] = _StdoutEofSentinel()
_ASYNC_TIMEOUT: Final[_AsyncTimeoutSentinel] = _AsyncTimeoutSentinel()
_ASYNC_CANCELLED: Final[_AsyncCancelledSentinel] = _AsyncCancelledSentinel()

_StdoutQueueItem: TypeAlias = str | _StdoutEofSentinel
_AsyncReadResult: TypeAlias = bytes | _AsyncTimeoutSentinel | _AsyncCancelledSentinel


@dataclass(frozen=True, slots=True)
class _SyncResources:
    stdin: IO[str]
    stdout: IO[str]
    stderr: IO[str]
    stdout_thread: threading.Thread
    stderr_thread: threading.Thread


class _SyncStdio(NamedTuple):
    stdin: IO[str]
    stdout: IO[str]
    stderr: IO[str]


class _SyncReaderThreads(NamedTuple):
    stdout_thread: threading.Thread
    stderr_thread: threading.Thread


class _AsyncStdio(NamedTuple):
    stdin: asyncio.StreamWriter
    stdout: asyncio.StreamReader
    stderr: asyncio.StreamReader


class CodexProcessRunnerBase(ABC):
    def __init__(self, *, executable_path: str) -> None:
        self._executable_path: Final[str] = executable_path

    @staticmethod
    def _check_cancelled(command: CodexExecCommand) -> None:
        signal = command.signal
        if signal is None:
            return

        if signal.is_set():
            raise CodexCancelledError("Turn cancelled")

    @staticmethod
    def _raise_exec_error(*, detail: str, stderr: str) -> NoReturn:
        raise CodexExecError(f"Codex Exec exited with {detail}", stderr=stderr)

    @abstractmethod
    def stream_lines(self, command: CodexExecCommand) -> Iterator[str] | AsyncIterator[str]:
        """Run command and stream stdout lines."""


class SyncCodexProcessRunner(CodexProcessRunnerBase):
    def stream_lines(self, command: CodexExecCommand) -> Iterator[str]:
        self._check_cancelled(command)

        process = self._spawn_process(command)
        stdio = self._require_stdio(process)

        stdout_queue: queue.Queue[_StdoutQueueItem] = queue.Queue()
        stderr_chunks: list[str] = []
        reader_threads = self._start_reader_threads(
            stdout=stdio.stdout,
            stderr=stdio.stderr,
            stdout_queue=stdout_queue,
            stderr_chunks=stderr_chunks,
        )
        resources = _SyncResources(
            stdin=stdio.stdin,
            stdout=stdio.stdout,
            stderr=stdio.stderr,
            stdout_thread=reader_threads.stdout_thread,
            stderr_thread=reader_threads.stderr_thread,
        )

        try:
            self._write_stdin(stdin=stdio.stdin, stdin_text=command.stdin)
            yield from self._iter_stdout_lines(
                process=process,
                command=command,
                stdout_queue=stdout_queue,
                stdout_thread=reader_threads.stdout_thread,
            )
            return_code = process.wait()
            reader_threads.stderr_thread.join()
            stderr_text = "".join(stderr_chunks)
            self._raise_on_bad_exit(return_code=return_code, stderr=stderr_text)
        except CodexCancelledError:
            self._terminate_process(process)
            raise
        finally:
            self._cleanup(process=process, resources=resources)

    def _spawn_process(self, command: CodexExecCommand) -> subprocess.Popen[str]:
        try:
            return subprocess.Popen(  # noqa: S603
                [self._executable_path, *command.argv],
                env=command.env,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                encoding="utf-8",
                errors="replace",
                bufsize=1,
            )
        except OSError as error:
            self._raise_exec_error(detail=f"spawn failure: {error}", stderr="")

    def _require_stdio(self, process: subprocess.Popen[str]) -> _SyncStdio:
        stdin = process.stdin
        stdout = process.stdout
        stderr = process.stderr
        if stdin is None or stdout is None or stderr is None:
            self._terminate_process(process)
            self._raise_exec_error(detail="spawn failure: missing stdio streams", stderr="")

        return _SyncStdio(stdin=stdin, stdout=stdout, stderr=stderr)

    def _start_reader_threads(
        self,
        *,
        stdout: IO[str],
        stderr: IO[str],
        stdout_queue: queue.Queue[_StdoutQueueItem],
        stderr_chunks: list[str],
    ) -> _SyncReaderThreads:
        stdout_thread = threading.Thread(
            target=self._pump_stdout,
            args=(stdout, stdout_queue),
            name="acodex-sync-stdout-reader",
            daemon=True,
        )
        stderr_thread = threading.Thread(
            target=self._pump_stderr,
            args=(stderr, stderr_chunks),
            name="acodex-sync-stderr-reader",
            daemon=True,
        )
        stdout_thread.start()
        stderr_thread.start()
        return _SyncReaderThreads(stdout_thread=stdout_thread, stderr_thread=stderr_thread)

    @staticmethod
    def _write_stdin(*, stdin: IO[str], stdin_text: str) -> None:
        with suppress(BrokenPipeError, OSError):
            stdin.write(stdin_text)
            stdin.flush()
        with suppress(OSError):
            stdin.close()

    def _iter_stdout_lines(
        self,
        *,
        process: subprocess.Popen[str],
        command: CodexExecCommand,
        stdout_queue: queue.Queue[_StdoutQueueItem],
        stdout_thread: threading.Thread,
    ) -> Iterator[str]:
        while True:
            self._check_cancelled(command)

            try:
                item = stdout_queue.get(timeout=_READ_POLL_INTERVAL_SECONDS)
            except queue.Empty:
                if process.poll() is not None and not stdout_thread.is_alive():
                    return
                continue

            if item is _STDOUT_EOF:
                return

            yield cast("str", item).rstrip("\r\n")

    @classmethod
    def _raise_on_bad_exit(cls, *, return_code: int, stderr: str) -> None:
        if return_code != 0:
            cls._raise_exec_error(detail=f"code {return_code}", stderr=stderr)

    def _cleanup(self, *, process: subprocess.Popen[str], resources: _SyncResources) -> None:
        self._terminate_process(process)
        with suppress(OSError):
            resources.stdin.close()
        with suppress(OSError):
            resources.stdout.close()
        with suppress(OSError):
            resources.stderr.close()
        resources.stdout_thread.join()
        resources.stderr_thread.join()

    @staticmethod
    def _pump_stdout(stream: IO[str], output: queue.Queue[_StdoutQueueItem]) -> None:
        try:
            for line in stream:
                output.put(line)
        finally:
            output.put(_STDOUT_EOF)

    @staticmethod
    def _pump_stderr(stream: IO[str], chunks: list[str]) -> None:
        chunks.append(stream.read())

    @staticmethod
    def _terminate_process(process: subprocess.Popen[str]) -> None:
        if process.poll() is not None:
            return

        with suppress(OSError):
            process.terminate()

        with suppress(subprocess.TimeoutExpired):
            process.wait(timeout=_PROCESS_TERMINATE_TIMEOUT_SECONDS)

        if process.poll() is not None:
            return

        with suppress(OSError):
            process.kill()
        with suppress(subprocess.TimeoutExpired):
            process.wait(timeout=_PROCESS_TERMINATE_TIMEOUT_SECONDS)


class AsyncCodexProcessRunner(CodexProcessRunnerBase):
    async def stream_lines(self, command: CodexExecCommand) -> AsyncIterator[str]:
        self._check_cancelled(command)

        process = await self._spawn_process(command)
        stdio: _AsyncStdio | None = None
        stderr_task: asyncio.Task[bytes] | None = None

        try:
            stdio = self._require_stdio(process)
            stderr_task = asyncio.create_task(
                stdio.stderr.read(),
                name="acodex-async-stderr-reader",
            )

            await self._write_stdin(stdin=stdio.stdin, stdin_text=command.stdin)
            async for line in self._iter_stdout_lines(command=command, stdout=stdio.stdout):
                yield line

            return_code = await process.wait()
            stderr_text = await self._decode_stderr(stderr_task)

            self._raise_on_bad_exit(return_code=return_code, stderr=stderr_text)
        except CodexCancelledError:
            await self._terminate_process(process)
            raise
        finally:
            await self._cleanup(
                process=process,
                stdin=stdio.stdin if stdio is not None else None,
                stderr_task=stderr_task,
            )

    async def _spawn_process(self, command: CodexExecCommand) -> asyncio.subprocess.Process:
        try:
            return await asyncio.create_subprocess_exec(
                self._executable_path,
                *command.argv,
                env=command.env,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except OSError as error:
            self._raise_exec_error(detail=f"spawn failure: {error}", stderr="")

    async def _cleanup(
        self,
        *,
        process: asyncio.subprocess.Process,
        stdin: asyncio.StreamWriter | None,
        stderr_task: asyncio.Task[bytes] | None,
    ) -> None:
        await self._terminate_process(process)

        if stdin is not None:
            if not stdin.is_closing():
                stdin.close()
            with suppress(BrokenPipeError, ConnectionResetError):
                await stdin.wait_closed()

        if stderr_task is not None:
            if not stderr_task.done():
                stderr_task.cancel()
            with suppress(asyncio.CancelledError):
                await stderr_task

    @staticmethod
    def _require_stdio(
        process: asyncio.subprocess.Process,
    ) -> _AsyncStdio:
        stdin = process.stdin
        stdout = process.stdout
        stderr = process.stderr
        if stdin is None or stdout is None or stderr is None:
            raise CodexExecError(
                "Codex Exec exited with spawn failure: missing stdio streams",
                stderr="",
            )

        return _AsyncStdio(stdin=stdin, stdout=stdout, stderr=stderr)

    @staticmethod
    async def _write_stdin(*, stdin: asyncio.StreamWriter, stdin_text: str) -> None:
        with suppress(BrokenPipeError, ConnectionResetError):
            stdin.write(stdin_text.encode("utf-8"))
            await stdin.drain()
        stdin.close()
        with suppress(BrokenPipeError, ConnectionResetError):
            await stdin.wait_closed()

    async def _iter_stdout_lines(
        self,
        *,
        command: CodexExecCommand,
        stdout: asyncio.StreamReader,
    ) -> AsyncIterator[str]:
        while True:
            self._check_cancelled(command)

            line = await self._read_next_line(command=command, stdout=stdout)
            if line is _ASYNC_TIMEOUT:
                continue
            if line is _ASYNC_CANCELLED:
                self._raise_cancelled()
            if line == b"":
                return

            yield cast("bytes", line).decode("utf-8", errors="replace").rstrip("\r\n")

    @staticmethod
    async def _read_next_line(
        *,
        command: CodexExecCommand,
        stdout: asyncio.StreamReader,
    ) -> _AsyncReadResult:
        signal = command.signal
        if isinstance(signal, asyncio.Event):
            readline_task = asyncio.create_task(
                stdout.readline(),
                name="acodex-async-stdout-readline",
            )
            cancelled_task = asyncio.create_task(
                signal.wait(),
                name="acodex-async-cancel-wait",
            )
            done, pending = await asyncio.wait(
                {readline_task, cancelled_task},
                return_when=asyncio.FIRST_COMPLETED,
            )
            for task in pending:
                task.cancel()
            if pending:
                await asyncio.gather(*pending, return_exceptions=True)

            if cancelled_task in done:
                return _ASYNC_CANCELLED

            return readline_task.result()

        try:
            return await asyncio.wait_for(
                stdout.readline(),
                timeout=_READ_POLL_INTERVAL_SECONDS,
            )
        except asyncio.TimeoutError:
            return _ASYNC_TIMEOUT

    @staticmethod
    def _raise_cancelled() -> None:
        raise CodexCancelledError("Turn cancelled")

    @classmethod
    def _raise_on_bad_exit(cls, *, return_code: int, stderr: str) -> None:
        if return_code != 0:
            cls._raise_exec_error(detail=f"code {return_code}", stderr=stderr)

    @staticmethod
    async def _decode_stderr(stderr_task: asyncio.Task[bytes]) -> str:
        return (await stderr_task).decode("utf-8", errors="replace")

    @staticmethod
    async def _terminate_process(process: asyncio.subprocess.Process) -> None:
        if process.returncode is not None:
            return

        with suppress(ProcessLookupError):
            process.terminate()

        with suppress(asyncio.TimeoutError):
            await asyncio.wait_for(process.wait(), timeout=_PROCESS_TERMINATE_TIMEOUT_SECONDS)

        with suppress(ProcessLookupError):
            process.kill()
        with suppress(asyncio.TimeoutError):
            await asyncio.wait_for(process.wait(), timeout=_PROCESS_TERMINATE_TIMEOUT_SECONDS)



from __future__ import annotations

import asyncio
import json
from pathlib import Path
from typing import cast

from acodex.codex import AsyncCodex, Codex
from acodex.types.input import UserInputLocalImage, UserInputText
from tests.unit.fake_codex_executable import create_fake_codex_executable


def test_codex_start_thread_runs_with_fake_executable(tmp_path: Path) -> None:
    executable = create_fake_codex_executable(tmp_path)
    client = Codex(
        codex_path_override=str(executable),
        env={
            "FAKE_CODEX_MODE": "thread_success",
            "FAKE_THREAD_ID": "codex-thread-sync",
            "FAKE_RESPONSES_JSON": json.dumps(["sync response"]),
        },
    )

    thread = client.start_thread()
    turn = thread.run("hello")

    assert thread.id == "codex-thread-sync"
    assert turn.final_response == "sync response"


def test_codex_resume_thread_forwards_resume_id(tmp_path: Path) -> None:
    executable = create_fake_codex_executable(tmp_path)
    client = Codex(
        codex_path_override=str(executable),
        env={
            "FAKE_CODEX_MODE": "normalize_capture",
        },
    )

    thread = client.resume_thread("thread-xyz")
    turn = thread.run(
        [
            UserInputText(text="hello"),
            UserInputLocalImage(path="/images/a.png"),
        ],
    )
    payload = cast("dict[str, object]", json.loads(turn.final_response))
    argv = cast("list[str]", payload["argv"])

    assert payload["resume_id"] == "thread-xyz"
    assert argv[argv.index("resume") + 1] == "thread-xyz"
    assert argv.index("resume") < argv.index("--image")


def test_async_codex_start_thread_runs_with_fake_executable(tmp_path: Path) -> None:
    executable = create_fake_codex_executable(tmp_path)
    client = AsyncCodex(
        codex_path_override=str(executable),
        env={
            "FAKE_CODEX_MODE": "thread_success",
            "FAKE_THREAD_ID": "codex-thread-async",
            "FAKE_RESPONSES_JSON": json.dumps(["async response"]),
        },
    )

    async def run() -> tuple[str | None, str]:
        thread = client.start_thread()
        turn = await thread.run("hello")
        return thread.id, turn.final_response

    thread_id, final_response = asyncio.run(run())

    assert thread_id == "codex-thread-async"
    assert final_response == "async response"


def test_async_codex_resume_thread_forwards_resume_id(tmp_path: Path) -> None:
    executable = create_fake_codex_executable(tmp_path)
    client = AsyncCodex(
        codex_path_override=str(executable),
        env={
            "FAKE_CODEX_MODE": "normalize_capture",
        },
    )

    async def run() -> dict[str, object]:
        thread = client.resume_thread("thread-xyz")
        turn = await thread.run(
            [
                UserInputText(text="hello"),
                UserInputLocalImage(path="/images/a.png"),
            ],
        )
        return cast("dict[str, object]", json.loads(turn.final_response))

    payload = asyncio.run(run())
    argv = cast("list[str]", payload["argv"])

    assert payload["resume_id"] == "thread-xyz"
    assert argv[argv.index("resume") + 1] == "thread-xyz"
    assert argv.index("resume") < argv.index("--image")
